import attr
from .._common import attrs_default
from .. import _exception, _session
from typing import Iterable, Sequence


@attrs_default
class PollOption:
    """Represents a poll option."""

    #: ID of the poll option
    id: str = attr.ib(converter=str)
    #: Text of the poll option
    text: str
    #: Whether vote when creating or client voted
    vote: bool
    #: ID of the users who voted for this poll option
    voters: Sequence[str]
    #: Votes count
    votes_count: int

    @classmethod
    def _from_graphql(cls, data):
        if data.get("viewer_has_voted") is None:
            vote = False
        elif isinstance(data["viewer_has_voted"], bool):
            vote = data["viewer_has_voted"]
        else:
            vote = data["viewer_has_voted"] == "true"
        return cls(
            id=int(data["id"]),
            text=data.get("text"),
            vote=vote,
            voters=(
                [m["node"]["id"] for m in data["voters"]["edges"]]
                if isinstance(data.get("voters"), dict)
                else data["voters"]
            ),
            votes_count=(
                data["voters"]["count"]
                if isinstance(data.get("voters"), dict)
                else data["total_count"]
            ),
        )


@attrs_default
class Poll:
    """Represents a poll."""

    #: ID of the poll
    session: _session.Session
    #: ID of the poll
    id: str = attr.ib(converter=str)
    #: The poll's question
    question: str
    #: The poll's top few options. The full list can be fetched with `fetch_options`
    options: Sequence[PollOption]
    #: Options count
    options_count: int

    @classmethod
    def _from_graphql(cls, session, data):
        return cls(
            session=session,
            id=data["id"],
            question=data["title"] if data.get("title") else data["text"],
            options=[PollOption._from_graphql(m) for m in data["options"]],
            options_count=data["total_count"],
        )

    async def fetch_options(self) -> Sequence[PollOption]:
        """Fetch all `PollOption` objects on the poll.

        The result is ordered with options with the most votes first.

        Example:
            >>> options = poll.fetch_options()
            >>> options[0].text
            "An option"
        """
        data = {"question_id": self.id}
        j = await self.session._payload_post("/ajax/mercury/get_poll_options", data)
        return [PollOption._from_graphql(m) for m in j]

    async def set_votes(self, option_ids: Iterable[str], new_options: Iterable[str] = None):
        """Update the user's poll vote.

        Args:
            option_ids: Option ids to vote for / keep voting for
            new_options: New options to add

        Example:
            >>> options = poll.fetch_options()
            >>> # Add option
            >>> poll.set_votes([o.id for o in options], new_options=["New option"])
            >>> # Remove vote from option
            >>> poll.set_votes([o.id for o in options if o.text != "Option 1"])
        """
        data = {"question_id": self.id}

        for i, option_id in enumerate(option_ids or ()):
            data["selected_options[{}]".format(i)] = option_id

        for i, option_text in enumerate(new_options or ()):
            data["new_options[{}]".format(i)] = option_text

        j = await self.session._payload_post(
            "/messaging/group_polling/update_vote/?dpr=1", data
        )
        if j.get("status") != "success":
            raise _exception.ExternalError(
                "Failed updating poll vote: {}".format(j.get("errorTitle")),
                j.get("errorMessage"),
            )



import attr
import time
import random
import paho.mqtt.client
import asyncio
import aiohttp
from ._common import log, kw_only
from . import _util, _exception, _session, _graphql, _events

from typing import AsyncGenerator, Optional, List


HOST = "edge-chat.messenger.com"

TOPICS = [
    # Things that happen in chats (e.g. messages)
    "/t_ms",
    # Group typing notifications
    "/thread_typing",
    # Private chat typing notifications
    "/orca_typing_notifications",
    # Active notifications
    "/orca_presence",
    # Other notifications not related to chats (e.g. friend requests)
    "/legacy_web",
    # Facebook's continuous error reporting/logging?
    "/br_sr",
    # Response to /br_sr
    "/sr_res",
    # Data about user-to-user calls
    # TODO: Investigate the response from this! (A bunch of binary data)
    # "/t_rtc",
    # TODO: Find out what this does!
    # TODO: Investigate the response from this! (A bunch of binary data)
    # "/t_p",
    # TODO: Find out what this does!
    "/webrtc",
    # TODO: Find out what this does!
    "/onevc",
    # TODO: Find out what this does!
    "/notify_disconnect",
    # Old, no longer active topics
    # These are here just in case something interesting pops up
    "/inbox",
    "/mercury",
    "/messaging_events",
    "/orca_message_notifications",
    "/pp",
    "/webrtc_response",
]


def get_cookie_header(session: aiohttp.ClientSession, url: str) -> str:
    """Extract a cookie header from a requests session."""
    # The cookies are extracted this way to make sure they're escaped correctly
    return session.cookie_jar.filter_cookies(url).output(header="", sep=";").lstrip()


def generate_session_id() -> int:
    """Generate a random session ID between 1 and 9007199254740991."""
    return random.randint(1, 2 ** 53)


def mqtt_factory() -> paho.mqtt.client.Client:
    # Configure internal MQTT handler
    mqtt = paho.mqtt.client.Client(
        client_id="mqttwsclient",
        clean_session=True,
        protocol=paho.mqtt.client.MQTTv31,
        transport="websockets",
    )
    mqtt.enable_logger()
    # mqtt.max_inflight_messages_set(20)  # The rest will get queued
    # mqtt.max_queued_messages_set(0)  # Unlimited messages can be queued
    # mqtt.message_retry_set(20)  # Retry sending for at least 20 seconds
    # mqtt.reconnect_delay_set(min_delay=1, max_delay=120)
    mqtt.tls_set()
    mqtt.connect_async(HOST, 443, keepalive=10)
    return mqtt


async def fetch_sequence_id(session: _session.Session) -> int:
    """Fetch sequence ID."""
    params = {
        "limit": 0,
        "tags": ["INBOX"],
        "before": None,
        "includeDeliveryReceipts": False,
        "includeSeqID": True,
    }
    log.debug("Fetching MQTT sequence ID")
    # Same doc id as in `Client.fetch_threads`
    (j,) = await session._graphql_requests(_graphql.from_doc_id("1349387578499440", params))
    sequence_id = j["viewer"]["message_threads"]["sync_sequence_id"]
    if not sequence_id:
        raise _exception.NotLoggedIn("Failed fetching sequence id")
    return int(sequence_id)


@attr.s(slots=True, kw_only=kw_only, eq=False, auto_attribs=True)
class Listener:
    """Listen to incoming Facebook events.

    Initialize a connection to the Facebook MQTT service.

    Args:
        session: The session to use when making requests.
        chat_on: Whether ...
        foreground: Whether ...

    Example:
        >>> listener = fbchat.Listener(session, chat_on=True, foreground=True)
    """

    session: _session.Session
    _chat_on: bool
    _foreground: bool
    _loop: asyncio.AbstractEventLoop = attr.ib(factory=asyncio.get_event_loop)
    _mqtt: paho.mqtt.client.Client = attr.ib(factory=mqtt_factory)
    _disconnect_error: Optional[Exception] = None
    _sync_token: Optional[str] = None
    _sequence_id: Optional[int] = None
    _tmp_events: List[_events.Event] = attr.ib(factory=list)
    _message_queue: asyncio.Queue = attr.ib(factory=lambda: asyncio.Queue(maxsize=64))

    def __attrs_post_init__(self):
        # Configure callbacks
        self._mqtt.on_message = self._on_message_handler
        self._mqtt.on_connect = self._on_connect_handler
        self._mqtt.on_socket_open = self.on_socket_open
        self._mqtt.on_socket_close = self.on_socket_close
        self._mqtt.on_socket_register_write = self.on_socket_register_write
        self._mqtt.on_socket_unregister_write = self.on_socket_unregister_write

    def on_socket_open(self, client, userdata, sock):
        self._loop.add_reader(sock, client.loop_read)

    def on_socket_close(self, client, userdata, sock):
        self._loop.remove_reader(sock)

    def on_socket_register_write(self, client, userdata, sock):
        self._loop.add_writer(sock, client.loop_write)

    def on_socket_unregister_write(self, client, userdata, sock):
        self._loop.remove_writer(sock)

    def _handle_ms(self, j):
        """Handle /t_ms special logic.

        Returns whether to continue parsing the message.
        """
        # TODO: Merge this with the parsing in _events

        # Update sync_token when received
        # This is received in the first message after we've created a messenger
        # sync queue.
        if "syncToken" in j and "firstDeltaSeqId" in j:
            self._sync_token = j["syncToken"]
            self._sequence_id = j["firstDeltaSeqId"]
            return False

        if "errorCode" in j:
            error = j["errorCode"]
            # TODO: 'F\xfa\x84\x8c\x85\xf8\xbc-\x88 FB_PAGES_INSUFFICIENT_PERMISSION\x00'
            if error in ("ERROR_QUEUE_NOT_FOUND", "ERROR_QUEUE_OVERFLOW"):
                # ERROR_QUEUE_NOT_FOUND means that the queue was deleted, since too
                # much time passed, or that it was simply missing
                # ERROR_QUEUE_OVERFLOW means that the sequence id was too small, so
                # the desired events could not be retrieved
                log.warning(
                    "The MQTT listener was disconnected for too long,"
                    " events may have been lost"
                )
                self._sync_token = None
                self._sequence_id = None
                return False
            log.error("MQTT error code %s received", error)
            return False

        # Update last sequence id
        # Except for the two cases above, this is always received
        self._sequence_id = j["lastIssuedSeqId"]
        return True

    def _on_message_handler(self, client, userdata, message):
        # Parse payload JSON
        try:
            j = _util.parse_json(message.payload.decode("utf-8"))
        except (_exception.FacebookError, UnicodeDecodeError):
            log.debug(message.payload)
            log.exception("Failed parsing MQTT data on %s as JSON", message.topic)
            return

        log.debug("MQTT payload: %s, %s", message.topic, j)

        if message.topic == "/t_ms":
            if not self._handle_ms(j):
                return

        try:
            for event in _events.parse_events(self.session, message.topic, j):
                self._message_queue.put_nowait(event)
        except _exception.ParseError:
            log.exception("Failed parsing MQTT data")

    def _on_connect_handler(self, client, userdata, flags, rc):
        if rc == 21:
            log.info("Return code 21 in connect handler, disconnecting and throwing error")
            self.disconnect()
            self._disconnect_error = _exception.NotConnected(
                "Failed connecting. Maybe your cookies are wrong?"
            )
            return
        if rc != 0:
            err = paho.mqtt.client.connack_string(rc)
            log.error("MQTT Connection Error: %s", err)
            return  # Don't try to send publish if the connection failed

        self._messenger_queue_publish()

    def _messenger_queue_publish(self):
        # configure receiving messages.
        payload = {
            "sync_api_version": 10,
            "max_deltas_able_to_process": 1000,
            "delta_batch_size": 500,
            "encoding": "JSON",
            "entity_fbid": self.session.user.id,
        }

        # If we don't have a sync_token, create a new messenger queue
        # This is done so that across reconnects, if we've received a sync token, we
        # SHOULD receive a piece of data in /t_ms exactly once!
        if self._sync_token is None:
            topic = "/messenger_sync_create_queue"
            payload["initial_titan_sequence_id"] = str(self._sequence_id)
            payload["device_params"] = None
        else:
            topic = "/messenger_sync_get_diffs"
            payload["last_seq_id"] = str(self._sequence_id)
            payload["sync_token"] = self._sync_token

        self._mqtt.publish(topic, _util.json_minimal(payload), qos=1)

    def _configure_connect_options(self):
        # Generate a new session ID on each reconnect
        session_id = generate_session_id()

        username = {
            # The user ID
            "u": self.session.user.id,
            # Session ID
            "s": session_id,
            # Active status setting
            "chat_on": self._chat_on,
            # foreground_state - Whether the window is focused
            "fg": self._foreground,
            # Can be any random ID
            "d": self.session._client_id,
            # Application ID, taken from facebook.com
            "aid": 219994525426954,
            # MQTT extension by FB, allows making a SUBSCRIBE while CONNECTing
            "st": TOPICS,
            # MQTT extension by FB, allows making a PUBLISH while CONNECTing
            # Using this is more efficient, but the same can be acheived with:
            #     def on_connect(*args):
            #         mqtt.publish(topic, payload, qos=1)
            #     mqtt.on_connect = on_connect
            # TODO: For some reason this doesn't work!
            "pm": [
                # {
                #     "topic": topic,
                #     "payload": payload,
                #     "qos": 1,
                #     "messageId": 65536,
                # }
            ],
            # Unknown parameters
            "cp": 3,
            "ecp": 10,
            "ct": "websocket",
            "mqtt_sid": "",
            "dc": "",
            "no_auto_fg": True,
            "gas": None,
            "pack": [],
        }

        self._mqtt.username_pw_set(_util.json_minimal(username))

        headers = {
            "Cookie": get_cookie_header(
                self.session._session, "https://edge-chat.messenger.com/chat"
            ),
            "User-Agent": self.session._session._default_headers["User-Agent"],
            "Origin": "https://www.messenger.com",
            "Host": HOST,
        }

        # TODO: Is region (lla | atn | odn | others?) important?
        self._mqtt.ws_set_options(
            path="/chat?sid={}".format(session_id), headers=headers
        )

    async def _reconnect(self) -> None:
        # Try reconnecting
        self._configure_connect_options()
        try:
            self._mqtt.reconnect()
        except (
            # Taken from .loop_forever
            paho.mqtt.client.socket.error,
            OSError,
            paho.mqtt.client.WebsocketConnectionError,
        ) as e:
            raise _exception.NotLoggedIn("MQTT reconnection failed") from e

    async def listen(self) -> AsyncGenerator[_events.Event, Optional[bool]]:
        """Run the listening loop continually.

        This is a blocking call, that will yield events as they arrive.

        Example:
            Print events continually.

            >>> listener = Listener(session)
            >>> async for event in listener.listen():
            ...     print(event)
        """
        if self._sequence_id is None:
            self._sequence_id = await fetch_sequence_id(self.session)
        # TODO add some limits for all the infinite while loops

        await self._reconnect()
        yield _events.Connect()

        while True:
            try:
                await asyncio.sleep(1)
            except asyncio.CancelledError:
                self.disconnect()
                # this might not be necessary
                self._mqtt.loop_misc()
                break
            rc = self._mqtt.loop_misc()

            # The sequence ID was reset in _handle_ms
            if self._sequence_id is None:
                self._sequence_id = await fetch_sequence_id(self.session)
                self._messenger_queue_publish()
                yield _events.Resync()

            # If disconnect() has been called
            # Beware, internal API, may have to change this to something more stable!
            if self._mqtt._state == paho.mqtt.client.mqtt_cs_disconnecting:
                break  # Stop listening

            if rc != paho.mqtt.client.MQTT_ERR_SUCCESS:
                # If known/expected error
                if rc == paho.mqtt.client.MQTT_ERR_CONN_LOST:
                    yield _events.Disconnect(reason="Connection lost, retrying")
                elif rc == paho.mqtt.client.MQTT_ERR_NOMEM:
                    # This error is wrongly classified
                    # See https://github.com/eclipse/paho.mqtt.python/issues/340
                    yield _events.Disconnect(reason="Connection error, retrying")
                elif rc == paho.mqtt.client.MQTT_ERR_CONN_REFUSED:
                    raise _exception.NotLoggedIn("MQTT connection refused")
                else:
                    err = paho.mqtt.client.error_string(rc)
                    log.error("MQTT Error: %s", err)
                    yield _events.Disconnect(reason=f"MQTT Error: {err}, retrying")

                await self._reconnect()
                yield _events.Connect()

            while True:
                try:
                    yield self._message_queue.get_nowait()
                except asyncio.QueueEmpty:
                    break
        if self._disconnect_error:
            log.info("disconnect_error is set, raising and clearing variable")
            err = self._disconnect_error
            self._disconnect_error = None
            raise err

    def disconnect(self) -> None:
        """Disconnect the MQTT listener.

        Can be called while listening, which will stop the listening loop.

        The `Listener` object should not be used after this is called!

        Example:
            Stop the listener when receiving a message with the text "/stop"

            >>> for event in listener.listen():
            ...     if isinstance(event, fbchat.MessageEvent):
            ...         if event.message.text == "/stop":
            ...             listener.disconnect()  # Almost the same "break"
        """
        self._mqtt.disconnect()

    def set_foreground(self, value: bool) -> None:
        """Set the ``foreground`` value while listening."""
        # TODO: Document what this actually does!
        payload = _util.json_minimal({"foreground": value})
        info = self._mqtt.publish("/foreground_state", payload=payload, qos=1)
        self._foreground = value
        # TODO: We can't wait for this, since the loop is running within the same thread
        # info.wait_for_publish()

    def set_chat_on(self, value: bool) -> None:
        """Set the ``chat_on`` value while listening."""
        # TODO: Document what this actually does!
        # TODO: Is this the right request to make?
        data = {"make_user_available_when_in_foreground": value}
        payload = _util.json_minimal(data)
        info = self._mqtt.publish("/set_client_settings", payload=payload, qos=1)
        self._chat_on = value
        # TODO: We can't wait for this, since the loop is running within the same thread
        # info.wait_for_publish()

    # def send_additional_contacts(self, additional_contacts):
    #     payload = _util.json_minimal({"additional_contacts": additional_contacts})
    #     info = self._mqtt.publish("/send_additional_contacts", payload=payload, qos=1)
    #
    # def browser_close(self):
    #     info = self._mqtt.publish("/browser_close", payload=b"{}", qos=1)



import attr
import datetime
from ._common import attrs_event, UnknownEvent, ThreadEvent
from .. import _exception, _util, _threads, _models

from typing import Optional


@attrs_event
class ReactionEvent(ThreadEvent):
    """Somebody reacted to a message."""

    #: Message that the user reacted to
    message: "_models.Message"

    reaction: Optional[str]
    """The reaction.

    Not limited to the ones in `Message.react`.

    If ``None``, the reaction was removed.
    """

    @classmethod
    def _parse(cls, session, data):
        thread = cls._get_thread(session, data)
        return cls(
            author=_threads.User(session=session, id=str(data["userId"])),
            thread=thread,
            message=_models.Message(thread=thread, id=data["messageId"]),
            reaction=data["reaction"] if data["action"] == 0 else None,
        )


@attrs_event
class UserStatusEvent(ThreadEvent):
    #: Whether the user was blocked or unblocked
    blocked: bool

    @classmethod
    def _parse(cls, session, data):
        return cls(
            author=_threads.User(session=session, id=str(data["actorFbid"])),
            thread=cls._get_thread(session, data),
            blocked=not data["canViewerReply"],
        )


@attrs_event
class LiveLocationEvent(ThreadEvent):
    """Somebody sent live location info."""

    # TODO: This!

    @classmethod
    def _parse(cls, session, data):
        from . import _location

        thread = cls._get_thread(session, data)
        for location_data in data["messageLiveLocations"]:
            message = _models.Message(thread=thread, id=data["messageId"])
            author = _threads.User(session=session, id=str(location_data["senderId"]))
            location = _location.LiveLocationAttachment._from_pull(location_data)

        return None


@attrs_event
class UnsendEvent(ThreadEvent):
    """Somebody unsent a message (which deletes it for everyone)."""

    #: The unsent message
    message: "_models.Message"
    #: When the message was unsent
    at: datetime.datetime

    @classmethod
    def _parse(cls, session, data):
        thread = cls._get_thread(session, data)
        return cls(
            author=_threads.User(session=session, id=str(data["senderID"])),
            thread=thread,
            message=_models.Message(thread=thread, id=data["messageID"]),
            at=_util.millis_to_datetime(data["deletionTimestamp"]),
        )


@attrs_event
class MessageReplyEvent(ThreadEvent):
    """Somebody replied to a message."""

    #: The sent message
    message: "_models.MessageData"
    #: The message that was replied to
    replied_to: "_models.MessageData"

    @classmethod
    def _parse(cls, session, data):
        metadata = data["message"]["messageMetadata"]
        thread = cls._get_thread(session, metadata)
        return cls(
            author=_threads.User(session=session, id=str(metadata["actorFbId"])),
            thread=thread,
            message=_models.MessageData._from_reply(thread, data["message"]),
            replied_to=_models.MessageData._from_reply(
                thread, data["repliedToMessage"]
            ),
        )


def parse_client_delta(session, data):
    if "deltaMessageReaction" in data:
        return ReactionEvent._parse(session, data["deltaMessageReaction"])
    elif "deltaChangeViewerStatus" in data:
        # TODO: Parse all `reason`
        if data["deltaChangeViewerStatus"]["reason"] == 2:
            return UserStatusEvent._parse(session, data["deltaChangeViewerStatus"])
    elif "liveLocationData" in data:
        return LiveLocationEvent._parse(session, data["liveLocationData"])
    elif "deltaRecallMessageData" in data:
        return UnsendEvent._parse(session, data["deltaRecallMessageData"])
    elif "deltaMessageReply" in data:
        return MessageReplyEvent._parse(session, data["deltaMessageReply"])
    return UnknownEvent(source="client payload", data=data)


def parse_client_payloads(session, data):
    payload = _util.parse_json("".join(chr(z) for z in data["payload"]))

    try:
        for delta in payload["deltas"]:
            yield parse_client_delta(session, delta)
    except _exception.ParseError:
        raise
    except Exception as e:
        raise _exception.ParseError("Error parsing ClientPayload", data=payload) from e



import attr
import datetime
import aiohttp
import random
import re
import json
from yarl import URL
from http.cookies import SimpleCookie, BaseCookie

# TODO: Only import when required
# Or maybe just replace usage with `html.parser`?
import bs4

from ._common import log, kw_only
from . import _graphql, _util, _exception

from typing import Optional, Mapping, Callable, Any, Awaitable


SERVER_JS_DEFINE_REGEX = re.compile(r'\(require\("ServerJS"\)\)\(\).handle\(')
SERVER_JS_DEFINE_JSON_DECODER = json.JSONDecoder()


def parse_server_js_define(html: str) -> Mapping[str, Any]:
    """Parse ``ServerJSDefine`` entries from a HTML document."""
    # Find points where we should start parsing
    define_splits = SERVER_JS_DEFINE_REGEX.split(html)

    # TODO: Extract jsmods "require" and "define" from `bigPipe.onPageletArrive`?

    # Skip leading entry
    _, *define_splits = define_splits

    if not define_splits:
        raise _exception.ParseError("Could not find any ServerJSDefine", data=html)
    if len(define_splits) > 1:
        raise _exception.ParseError("Found too many ServerJSDefine", data=define_splits)
    try:
        parsed, _ = SERVER_JS_DEFINE_JSON_DECODER.raw_decode(define_splits[0], idx=0)
    except json.JSONDecodeError as e:
        raise _exception.ParseError("Invalid ServerJSDefine", data=define_splits[0]) from e
    try:
        rtn = parsed["define"]
    except KeyError:
        raise _exception.ParseError("Invalid ServerJSDefine", data=parsed)

    if not isinstance(rtn, list):
        raise _exception.ParseError("Invalid ServerJSDefine", data=rtn)

    # Convert to a dict
    return _util.get_jsmods_define(rtn)


def base36encode(number: int) -> str:
    """Convert from Base10 to Base36."""
    # Taken from https://en.wikipedia.org/wiki/Base36#Python_implementation
    chars = "0123456789abcdefghijklmnopqrstuvwxyz"

    sign = "-" if number < 0 else ""
    number = abs(number)
    result = ""

    while number > 0:
        number, remainder = divmod(number, 36)
        result = chars[remainder] + result

    return sign + result


def prefix_url(url: str) -> str:
    if url.startswith("/"):
        return "https://www.messenger.com" + url
    return url


def generate_message_id(now: datetime.datetime, client_id: str) -> str:
    k = _util.datetime_to_millis(now)
    l = int(random.random() * 4294967295)
    return "<{}:{}-{}@mail.projektitan.com>".format(k, l, client_id)


def get_user_id(session: aiohttp.ClientSession) -> str:
    try:
        rtn = session.cookie_jar.filter_cookies(URL("https://messenger.com")).get("c_user")
    except (AttributeError, KeyError):
        raise _exception.ParseError("Could not find user id", data=session.cookie_jar._cookies)
    if rtn is None:
        raise _exception.ParseError("Could not find user id", data=session.cookie_jar._cookies)
    return rtn if isinstance(rtn, str) else str(rtn.value)


def session_factory() -> aiohttp.ClientSession:
    from . import __version__

    session = aiohttp.ClientSession(headers={
        "Referer": "https://www.messenger.com/",
        # We won't try to set a fake user agent to mask our presence!
        # Facebook allows us access anyhow, and it makes our motives clearer:
        # We're not trying to cheat Facebook, we simply want to access their service
        "User-Agent": "fbchat-asyncio/{}".format(__version__)
    })
    return session


def client_id_factory() -> str:
    return hex(int(random.random() * 2 ** 31))[2:]


def find_form_request(html: str):
    soup = bs4.BeautifulSoup(html, "html.parser", parse_only=bs4.SoupStrainer("form"))

    form = soup.form
    if not form:
        raise _exception.ParseError("Could not find form to submit", data=html)

    url = form.get("action")
    if not url:
        raise _exception.ParseError("Could not find url to submit to", data=form)

    # From what I've seen, it'll always do this!
    if url.startswith("/"):
        url = "https://www.facebook.com" + url

    # It's okay to set missing values to something crap, the values are localized, and
    # hence are not available in the raw HTML
    data = {
        x["name"]: x.get("value", "[missing]")
        for x in form.find_all(["input", "button"])
    }
    return url, data


async def two_factor_helper(session: aiohttp.ClientSession, r: aiohttp.ClientResponse,
                            on_2fa_callback: Callable[[], Awaitable[int]]) -> str:
    url, data = find_form_request(await r.text())

    # You don't have to type a code if your device is already saved
    # Repeats if you get the code wrong
    while "approvals_code" in data:
        data["approvals_code"] = await on_2fa_callback()
        log.info("Submitting 2FA code")
        r = await session.post(url, data=data, allow_redirects=False)
        log.debug("2FA location: %s", r.headers.get("Location"))
        url, data = find_form_request(await r.text())

    # TODO: Can be missing if checkup flow was done on another device in the meantime?
    if "name_action_selected" in data:
        data["name_action_selected"] = "save_device"
        log.info("Saving browser")
        r = await session.post(url, data=data, allow_redirects=False)
        log.debug("2FA location: %s", r.headers.get("Location"))
        url, data = find_form_request(await r.text())

    log.info("Starting Facebook checkup flow")
    r = await session.post(url, data=data, allow_redirects=False)
    log.debug("2FA location: %s", r.headers.get("Location"))

    url, data = find_form_request(await r.text())
    if "submit[This was me]" not in data or "submit[This wasn't me]" not in data:
        raise _exception.ParseError("Could not fill out form properly (2)", data=data)
    data["submit[This was me]"] = "[any value]"
    del data["submit[This wasn't me]"]
    log.info("Verifying login attempt")

    r = await session.post(url, data=data, allow_redirects=False)
    log.debug("2FA location: %s", r.headers.get("Location"))

    url, data = find_form_request(await r.text())
    if "name_action_selected" not in data:
        raise _exception.ParseError("Could not fill out form properly (3)", data=data)
    data["name_action_selected"] = "save_device"
    log.info("Saving device again")

    r = await session.post(url, data=data, allow_redirects=False)
    log.debug("2FA location: %s", r.headers.get("Location"))
    return r.headers.get("Location")


def get_error_data(html: str) -> Optional[str]:
    """Get error message from a request."""
    soup = bs4.BeautifulSoup(
        html, "html.parser", parse_only=bs4.SoupStrainer("form", id="login_form")
    )
    # Attempt to extract and format the error string
    # The error message is in the user's own language!
    return " ".join(list(soup.stripped_strings)[1:3]) or None


def get_fb_dtsg(define) -> Optional[str]:
    if "DTSGInitData" in define:
        return define["DTSGInitData"]["token"]
    elif "DTSGInitialData" in define:
        return define["DTSGInitialData"]["token"]
    return None


@attr.s(slots=True, kw_only=kw_only, repr=False, eq=False, auto_attribs=True)
class Session:
    """Stores and manages state required for most Facebook requests.

    This is the main class, which is used to login to Facebook.
    """

    _user_id: str
    _fb_dtsg: str
    _revision: int
    _session: aiohttp.ClientSession = attr.ib(factory=session_factory)
    _counter: int = 0
    _client_id: str = attr.ib(factory=client_id_factory)

    @property
    def user(self):
        """The logged in user."""
        from . import _threads

        # TODO: Consider caching the result

        return _threads.User(session=self, id=self._user_id)

    def __repr__(self) -> str:
        # An alternative repr, to illustrate that you can't create the class directly
        return "<fbchat.Session user_id={}>".format(self._user_id)

    def _get_params(self):
        self._counter += 1  # TODO: Make this operation atomic / thread-safe
        return {
            "__a": 1,
            "__req": base36encode(self._counter),
            "__rev": self._revision,
            "fb_dtsg": self._fb_dtsg,
        }

    # TODO: Add ability to load previous cookies in here, to avoid 2fa flow
    @classmethod
    async def login(
        cls, email: str, password: str, on_2fa_callback: Callable[[], Awaitable[int]] = None
    ):
        """Login the user, using ``email`` and ``password``.

        Args:
            email: Facebook ``email``, ``id`` or ``phone number``
            password: Facebook account password
            on_2fa_callback: Function that will be called, in case a two factor
                authentication code is needed. This should return the requested code.

                Only tested using SMS, might not work with authentication applications.

                Note: Facebook limits the amount of codes they will give you, so if you
                don't receive a code, be patient, and try again later!

        Example:
            >>> import fbchat
            >>> import getpass
            >>> session = fbchat.Session.login(
            ...     input("Email: "),
            ...     getpass.getpass(),
            ...     on_2fa_callback=lambda: input("2FA Code: ")
            ... )
            Email: abc@gmail.com
            Password: ****
            2FA Code: 123456
            >>> session.user.id
            "1234"
        """
        session = session_factory()

        data = {
            # "jazoest": "2754",
            # "lsd": "AVqqqRUa",
            "initial_request_id": "x",  # any, just has to be present
            # "timezone": "-120",
            # "lgndim": "eyJ3IjoxNDQwLCJoIjo5MDAsImF3IjoxNDQwLCJhaCI6ODc3LCJjIjoyNH0=",
            # "lgnrnd": "044039_RGm9",
            "lgnjs": "n",
            "email": email,
            "pass": password,
            "login": "1",
            "persistent": "1",  # Changes the cookie type to have a long "expires"
            "default_persistent": "0",
        }

        try:
            # Should hit a redirect to https://www.messenger.com/
            # If this does happen, the session is logged in!
            r = await session.post(
                "https://www.messenger.com/login/password/",
                data=data,
                allow_redirects=False,
            )
        except aiohttp.ClientError as e:
            _exception.handle_requests_error(e)
            raise Exception("handle_requests_error did not raise exception")
        _exception.handle_http_error(r.status)

        url = r.headers.get("Location")

        # We weren't redirected, hence the email or password was wrong
        if not url:
            error = get_error_data(await r.text())
            raise _exception.NotLoggedIn(error)

        if "checkpoint" in url:
            if not on_2fa_callback:
                raise _exception.NotLoggedIn(
                    "2FA code required! Please supply `on_2fa_callback` to .login"
                )
            # Get a facebook.com/checkpoint/start url that handles the 2FA flow
            # This probably works differently for Messenger-only accounts
            url = _util.get_url_parameter(url, "next")
            if not url.startswith("https://www.facebook.com/checkpoint/start/"):
                raise _exception.ParseError("Failed 2fa flow (1)", data=url)

            r = await session.get(url, allow_redirects=False)
            url = r.headers.get("Location")
            if not url or not url.startswith("https://www.facebook.com/checkpoint/"):
                raise _exception.ParseError("Failed 2fa flow (2)", data=url)

            r = await session.get(url, allow_redirects=False)
            url = await two_factor_helper(session, r, on_2fa_callback)

            if not url.startswith("https://www.messenger.com/login/auth_token/"):
                raise _exception.ParseError("Failed 2fa flow (3)", data=url)

            r = await session.get(url, allow_redirects=False)
            url = r.headers.get("Location")

        if url != "https://www.messenger.com/":
            error = get_error_data(await r.text())
            raise _exception.NotLoggedIn("Failed logging in: {}, {}".format(url, error))

        try:
            return await cls._from_session(session=session)
        except _exception.NotLoggedIn as e:
            raise _exception.ParseError("Failed loading session", data=r) from e

    async def is_logged_in(self) -> bool:
        """Send a request to Facebook to check the login status.

        Returns:
            Whether the user is still logged in

        Example:
            >>> assert session.is_logged_in()
        """
        # Send a request to the login url, to see if we're directed to the home page
        try:
            r = await self._session.get(prefix_url("/login/"), allow_redirects=False)
        except aiohttp.ClientError as e:
            _exception.handle_requests_error(e)
            raise Exception("handle_requests_error did not raise exception")
        _exception.handle_http_error(r.status)
        return "https://www.messenger.com/" == r.headers.get("Location")

    async def logout(self) -> None:
        """Safely log out the user.

        The session object must not be used after this action has been performed!

        Example:
            >>> session.logout()
        """
        data = {"fb_dtsg": self._fb_dtsg}
        try:
            r = await self._session.post(
                prefix_url("/logout/"), data=data, allow_redirects=False
            )
        except aiohttp.ClientError as e:
            _exception.handle_requests_error(e)
            raise Exception("handle_requests_error did not raise exception")
        _exception.handle_http_error(r.status)

        if "Location" not in r.headers:
            raise _exception.FacebookError("Failed logging out, was not redirected!")
        if "https://www.messenger.com/login/" != r.headers["Location"]:
            raise _exception.FacebookError(
                "Failed logging out, got bad redirect: {}".format(r.headers["Location"])
            )

    @classmethod
    async def _from_session(cls, session):
        # TODO: Automatically set user_id when the cookie changes in the session
        user_id = get_user_id(session)

        # Make a request to the main page to retrieve ServerJSDefine entries
        try:
            r = await session.get(prefix_url("/"), allow_redirects=False)
        except aiohttp.ClientError as e:
            _exception.handle_requests_error(e)
            raise Exception("handle_requests_error did not raise exception")
        _exception.handle_http_error(r.status)

        define = parse_server_js_define(await r.text())

        fb_dtsg = get_fb_dtsg(define)
        if fb_dtsg is None:
            raise _exception.ParseError("Could not find fb_dtsg", data=define)
        if not fb_dtsg:
            # Happens when the client is not actually logged in
            raise _exception.NotLoggedIn(
                "Found empty fb_dtsg, the session was probably invalid."
            )

        try:
            revision = int(define["SiteData"]["client_revision"])
        except TypeError:
            raise _exception.ParseError("Could not find client revision", data=define)

        return cls(user_id=user_id, fb_dtsg=fb_dtsg, revision=revision, session=session)

    def get_cookies(self) -> Optional[Mapping[str, str]]:
        """Retrieve session cookies, that can later be used in `from_cookies`.

        Returns:
            A dictionary containing session cookies

        Example:
            >>> cookies = session.get_cookies()
        """
        cookie = self._session.cookie_jar.filter_cookies(URL("https://messenger.com"))
        return {key: morsel.value for key, morsel in cookie.items()}

    @classmethod
    async def from_cookies(cls, cookies: Mapping[str, str]):
        """Load a session from session cookies.

        Args:
            cookies: A dictionary containing session cookies

        Example:
            >>> cookies = session.get_cookies()
            >>> # Store cookies somewhere, and then subsequently
            >>> session = fbchat.Session.from_cookies(cookies)
        """
        session = session_factory()

        if isinstance(cookies, BaseCookie):
            cookie = cookies
        else:
            cookie = SimpleCookie()
            for key, value in cookies.items():
                cookie[key] = value
                cookie[key].update({"domain": "messenger.com", "path": "/"})
        session.cookie_jar.update_cookies(cookie, URL("https://messenger.com"))

        return await cls._from_session(session=session)

    async def _post(self, url, data, files=None, as_graphql=False):
        data.update(self._get_params())
        if files:
            payload = aiohttp.FormData()
            for key, value in data.items():
                payload.add_field(key, str(value))
            for key, (name, file, content_type) in files.items():
                payload.add_field(key, file, filename=name, content_type=content_type)
            data = payload
        try:
            r = await self._session.post(prefix_url(url), data=data)
        except aiohttp.ClientError as e:
            _exception.handle_requests_error(e)
            raise Exception("handle_requests_error did not raise exception")
        _exception.handle_http_error(r.status)
        text = await r.text()
        if text is None or len(text) == 0:
            raise _exception.HTTPError("Error when sending request: Got empty response")
        if as_graphql:
            return _graphql.response_to_json(text)
        else:
            text = _util.strip_json_cruft(text)
            j = _util.parse_json(text)
            log.debug(j)
            return j

    async def _payload_post(self, url, data, files=None):
        j = await self._post(url, data, files=files)
        _exception.handle_payload_error(j)

        # update fb_dtsg token if received in response
        if "jsmods" in j:
            define = _util.get_jsmods_define(j["jsmods"]["define"])
            fb_dtsg = get_fb_dtsg(define)
            if fb_dtsg:
                self._fb_dtsg = fb_dtsg

        try:
            return j["payload"]
        except (KeyError, TypeError) as e:
            raise _exception.ParseError("Missing payload", data=j) from e

    async def _graphql_requests(self, *queries):
        # TODO: Explain usage of GraphQL, probably in the docs
        # Perhaps provide this API as public?
        data = {
            "method": "GET",
            "response_format": "json",
            "queries": _graphql.queries_to_json(*queries),
        }
        return await self._post("/api/graphqlbatch/", data, as_graphql=True)

    async def _do_send_request(self, data):
        now = datetime.datetime.utcnow()
        offline_threading_id = _util.generate_offline_threading_id()
        data["client"] = "mercury"
        data["author"] = "fbid:{}".format(self._user_id)
        data["timestamp"] = _util.datetime_to_millis(now)
        data["source"] = "source:chat:web"
        data["offline_threading_id"] = offline_threading_id
        data["message_id"] = offline_threading_id
        data["threading_id"] = generate_message_id(now, self._client_id)
        data["ephemeral_ttl_mode:"] = "0"
        j = await self._post("/messaging/send/", data)

        _exception.handle_payload_error(j)

        try:
            message_ids = [
                (action["message_id"], action["thread_fbid"])
                for action in j["payload"]["actions"]
                if "message_id" in action
            ]
            if len(message_ids) != 1:
                log.warning("Got multiple message ids' back: {}".format(message_ids))
            return message_ids[0]
        except (KeyError, IndexError, TypeError) as e:
            raise _exception.ParseError("No message IDs could be found", data=j) from e



import attr
import datetime
import enum
from string import Formatter
from . import _attachment, _location, _file, _quick_reply, _sticker
from .._common import log, attrs_default
from .. import _exception, _util
from typing import Optional, Mapping, Sequence, Any, TYPE_CHECKING

if TYPE_CHECKING:
    from .. import _threads


class EmojiSize(enum.Enum):
    """Used to specify the size of a sent emoji."""

    LARGE = "369239383222810"
    MEDIUM = "369239343222814"
    SMALL = "369239263222822"

    @classmethod
    def _from_tags(cls, tags):
        string_to_emojisize = {
            "large": cls.LARGE,
            "medium": cls.MEDIUM,
            "small": cls.SMALL,
            "l": cls.LARGE,
            "m": cls.MEDIUM,
            "s": cls.SMALL,
        }
        for tag in tags or ():
            data = tag.split(":", 1)
            if len(data) > 1 and data[0] == "hot_emoji_size":
                return string_to_emojisize.get(data[1])
        return None


@attrs_default
class Mention:
    """Represents a ``@mention``.

    >>> fbchat.Mention(thread_id="1234", offset=5, length=2)
    Mention(thread_id="1234", offset=5, length=2)
    """

    #: The thread ID the mention is pointing at
    thread_id: str
    #: The character where the mention starts
    offset: int
    #: The length of the mention
    length: int

    @classmethod
    def _from_range(cls, data):
        # TODO: Parse data["entity"]["__typename"]
        return cls(
            # Can be missing
            thread_id=data["entity"].get("id"),
            offset=data["offset"],
            length=data["length"],
        )

    @classmethod
    def _from_prng(cls, data):
        return cls(thread_id=data["i"], offset=data["o"], length=data["l"])

    def _to_send_data(self, i):
        return {
            "profile_xmd[{}][id]".format(i): self.thread_id,
            "profile_xmd[{}][offset]".format(i): self.offset,
            "profile_xmd[{}][length]".format(i): self.length,
            "profile_xmd[{}][type]".format(i): "p",
        }


# Exaustively searched for options by using the list in:
# https://unicode.org/emoji/charts/full-emoji-list.html
SENDABLE_REACTIONS = ("❤", "😍", "😆", "😮", "😢", "😠", "👍", "👎")


@attrs_default
class Message:
    """Represents a Facebook message.

    Example:
        >>> thread = fbchat.User(session=session, id="1234")
        >>> message = fbchat.Message(thread=thread, id="mid.$XYZ")
    """

    #: The thread that this message belongs to.
    thread: '_threads.ThreadABC'
    #: The message ID.
    id: str = attr.ib(converter=str)

    @property
    def session(self):
        """The session to use when making requests."""
        return self.thread.session

    @staticmethod
    async def _delete_many(session, message_ids):
        data = {}
        for i, id_ in enumerate(message_ids):
            data["message_ids[{}]".format(i)] = id_
        j = await session._payload_post("/ajax/mercury/delete_messages.php?dpr=1", data)

    async def delete(self):
        """Delete the message (removes it only for the user).

        If you want to delete multiple messages, please use `Client.delete_messages`.

        Example:
            >>> message.delete()
        """
        await self._delete_many(self.session, [self.id])

    async def unsend(self):
        """Unsend the message (removes it for everyone).

        The message must to be sent by you, and less than 10 minutes ago.

        Example:
            >>> message.unsend()
        """
        data = {"message_id": self.id}
        j = await self.session._payload_post("/messaging/unsend_message/?dpr=1", data)

    async def react(self, reaction: Optional[str]):
        """React to the message, or removes reaction.

        Currently, you can use "❤", "😍", "😆", "😮", "😢", "😠", "👍" or "👎". It
        should be possible to add support for more, but we haven't figured that out yet.

        Args:
            reaction: Reaction emoji to use, or if ``None``, removes reaction.

        Example:
            >>> message.react("😍")
        """
        if reaction and reaction not in SENDABLE_REACTIONS:
            raise ValueError(
                "Invalid reaction! Please use one of: {}".format(SENDABLE_REACTIONS)
            )

        data = {
            "action": "ADD_REACTION" if reaction else "REMOVE_REACTION",
            "client_mutation_id": "1",
            "actor_id": self.session.user.id,
            "message_id": self.id,
            "reaction": reaction,
        }
        data = {
            "doc_id": 1491398900900362,
            "variables": _util.json_minimal({"data": data}),
        }
        j = await self.session._payload_post("/webgraphql/mutation", data)
        _exception.handle_graphql_errors(j)

    async def fetch(self) -> "MessageData":
        """Fetch fresh `MessageData` object.

        Example:
            >>> message = message.fetch()
            >>> message.text
            "The message text"
        """
        message_info = await self.thread._forced_fetch(self.id).get("message")
        return MessageData._from_graphql(self.thread, message_info)

    @staticmethod
    def format_mentions(text, *args, **kwargs):
        """Like `str.format`, but takes tuples with a thread id and text instead.

        Return a tuple, with the formatted string and relevant mentions.

        >>> Message.format_mentions("Hey {!r}! My name is {}", ("1234", "Peter"), ("4321", "Michael"))
        ("Hey 'Peter'! My name is Michael", [Mention(thread_id=1234, offset=4, length=7), Mention(thread_id=4321, offset=24, length=7)])

        >>> Message.format_mentions("Hey {p}! My name is {}", ("1234", "Michael"), p=("4321", "Peter"))
        ('Hey Peter! My name is Michael', [Mention(thread_id=4321, offset=4, length=5), Mention(thread_id=1234, offset=22, length=7)])
        """
        result = ""
        mentions = list()
        offset = 0
        f = Formatter()
        field_names = [field_name[1] for field_name in f.parse(text)]
        automatic = "" in field_names
        i = 0

        for (literal_text, field_name, format_spec, conversion) in f.parse(text):
            offset += len(literal_text)
            result += literal_text

            if field_name is None:
                continue

            if field_name == "":
                field_name = str(i)
                i += 1
            elif automatic and field_name.isdigit():
                raise ValueError(
                    "cannot switch from automatic field numbering to manual field specification"
                )

            thread_id, name = f.get_field(field_name, args, kwargs)[0]

            if format_spec:
                name = f.format_field(name, format_spec)
            if conversion:
                name = f.convert_field(name, conversion)

            result += name
            mentions.append(
                Mention(thread_id=thread_id, offset=offset, length=len(name))
            )
            offset += len(name)

        return result, mentions


@attrs_default
class MessageSnippet(Message):
    """Represents data in a Facebook message snippet.

    Inherits `Message`.
    """

    #: ID of the sender
    author: str
    #: When the message was sent
    created_at: datetime.datetime
    #: The actual message
    text: str
    #: A dict with offsets, mapped to the matched text
    matched_keywords: Mapping[int, str]

    @classmethod
    def _parse(cls, thread, data):
        return cls(
            thread=thread,
            id=data["message_id"],
            author=data["author"].rstrip("fbid:"),
            created_at=_util.millis_to_datetime(data["timestamp"]),
            text=data["body"],
            matched_keywords={int(k): v for k, v in data["matched_keywords"].items()},
        )


@attrs_default
class MessageData(Message):
    """Represents data in a Facebook message.

    Inherits `Message`.
    """

    #: ID of the sender
    author: str
    #: When the message was sent
    created_at: datetime.datetime
    #: The actual message
    text: Optional[str] = None
    #: A list of `Mention` objects
    mentions: Sequence[Mention] = attr.ib(factory=list)
    #: Size of a sent emoji
    emoji_size: Optional[EmojiSize] = None
    #: Whether the message is read
    is_read: Optional[bool] = None
    #: People IDs who read the message, only works with `ThreadABC.fetch_messages`
    read_by: bool = attr.ib(factory=list)
    #: A dictionary with user's IDs as keys, and their reaction as values
    reactions: Mapping[str, str] = attr.ib(factory=dict)
    #: A `Sticker`
    sticker: Optional[_sticker.Sticker] = None
    #: A list of attachments
    attachments: Sequence[_attachment.Attachment] = attr.ib(factory=list)
    #: A list of `QuickReply`
    quick_replies: Sequence[_quick_reply.QuickReply] = attr.ib(factory=list)
    #: Whether the message is unsent (deleted for everyone)
    unsent: Optional[bool] = False
    #: Message ID you want to reply to
    reply_to_id: Optional[str] = None
    #: Replied message
    replied_to: Optional[Any] = None
    #: Whether the message was forwarded
    forwarded: Optional[bool] = False

    @staticmethod
    def _get_forwarded_from_tags(tags):
        if tags is None:
            return False
        return any(map(lambda tag: "forward" in tag or "copy" in tag, tags))

    @staticmethod
    def _parse_quick_replies(data):
        if data:
            data = _util.parse_json(data).get("quick_replies")
            if isinstance(data, list):
                return [_quick_reply.graphql_to_quick_reply(q) for q in data]
            elif isinstance(data, dict):
                return [_quick_reply.graphql_to_quick_reply(data, is_response=True)]
        return []

    @classmethod
    def _from_graphql(cls, thread, data, read_receipts=None):
        if data.get("message_sender") is None:
            data["message_sender"] = {}
        if data.get("message") is None:
            data["message"] = {}
        tags = data.get("tags_list")

        created_at = _util.millis_to_datetime(int(data.get("timestamp_precise")))

        attachments = [
            _file.graphql_to_attachment(attachment)
            for attachment in data.get("blob_attachments") or ()
        ]
        unsent = False
        if data.get("extensible_attachment") is not None:
            attachment = graphql_to_extensible_attachment(data["extensible_attachment"])
            if isinstance(attachment, _attachment.UnsentMessage):
                unsent = True
            elif attachment:
                attachments.append(attachment)

        replied_to = None
        if data.get("replied_to_message") and data["replied_to_message"]["message"]:
            # data["replied_to_message"]["message"] is None if the message is deleted
            replied_to = cls._from_graphql(
                thread, data["replied_to_message"]["message"]
            )

        return cls(
            thread=thread,
            id=str(data["message_id"]),
            author=str(data["message_sender"]["id"]),
            created_at=created_at,
            text=data["message"].get("text"),
            mentions=[
                Mention._from_range(m) for m in data["message"].get("ranges") or ()
            ],
            emoji_size=EmojiSize._from_tags(tags),
            is_read=not data["unread"] if data.get("unread") is not None else None,
            read_by=[
                receipt["actor"]["id"]
                for receipt in read_receipts or ()
                if _util.millis_to_datetime(int(receipt["watermark"])) >= created_at
            ],
            reactions={
                str(r["user"]["id"]): r["reaction"] for r in data["message_reactions"]
            },
            sticker=_sticker.Sticker._from_graphql(data.get("sticker")),
            attachments=attachments,
            quick_replies=cls._parse_quick_replies(data.get("platform_xmd_encoded")),
            unsent=unsent,
            reply_to_id=replied_to.id if replied_to else None,
            replied_to=replied_to,
            forwarded=cls._get_forwarded_from_tags(tags),
        )

    @classmethod
    def _from_reply(cls, thread, data):
        tags = data["messageMetadata"].get("tags")
        metadata = data.get("messageMetadata", {})

        attachments = []
        unsent = False
        sticker = None
        for attachment in data.get("attachments") or ():
            attachment = _util.parse_json(attachment["mercuryJSON"])
            if attachment.get("blob_attachment"):
                attachments.append(
                    _file.graphql_to_attachment(attachment["blob_attachment"])
                )
            if attachment.get("extensible_attachment"):
                extensible_attachment = graphql_to_extensible_attachment(
                    attachment["extensible_attachment"]
                )
                if isinstance(extensible_attachment, _attachment.UnsentMessage):
                    unsent = True
                else:
                    attachments.append(extensible_attachment)
            if attachment.get("sticker_attachment"):
                sticker = _sticker.Sticker._from_graphql(
                    attachment["sticker_attachment"]
                )

        return cls(
            thread=thread,
            id=metadata.get("messageId"),
            author=str(metadata["actorFbId"]),
            created_at=_util.millis_to_datetime(metadata["timestamp"]),
            text=data.get("body"),
            mentions=[
                Mention._from_prng(m)
                for m in _util.parse_json(data.get("data", {}).get("prng", "[]"))
            ],
            emoji_size=EmojiSize._from_tags(tags),
            sticker=sticker,
            attachments=attachments,
            quick_replies=cls._parse_quick_replies(data.get("platform_xmd_encoded")),
            unsent=unsent,
            reply_to_id=data["messageReply"]["replyToMessageId"]["id"]
            if "messageReply" in data
            else None,
            forwarded=cls._get_forwarded_from_tags(tags),
        )

    @classmethod
    def _from_pull(cls, thread, data, author, created_at):
        metadata = data["messageMetadata"]

        tags = metadata.get("tags")

        mentions = []
        if data.get("data") and data["data"].get("prng"):
            try:
                mentions = [
                    Mention._from_prng(m)
                    for m in _util.parse_json(data["data"]["prng"])
                ]
            except Exception:
                log.exception("An exception occured while reading attachments")

        attachments = []
        unsent = False
        sticker = None
        try:
            for a in data.get("attachments") or ():
                mercury = a["mercury"]
                if mercury.get("blob_attachment"):
                    image_metadata = a.get("imageMetadata", {})
                    attach_type = mercury["blob_attachment"]["__typename"]
                    attachment = _file.graphql_to_attachment(
                        mercury["blob_attachment"], a.get("fileSize")
                    )
                    attachments.append(attachment)

                elif mercury.get("sticker_attachment"):
                    sticker = _sticker.Sticker._from_graphql(
                        mercury["sticker_attachment"]
                    )

                elif mercury.get("extensible_attachment"):
                    attachment = graphql_to_extensible_attachment(
                        mercury["extensible_attachment"]
                    )
                    if isinstance(attachment, _attachment.UnsentMessage):
                        unsent = True
                    elif attachment:
                        attachments.append(attachment)

        except Exception:
            log.exception(
                "An exception occured while reading attachments: {}".format(
                    data["attachments"]
                )
            )

        return cls(
            thread=thread,
            id=metadata["messageId"],
            author=author,
            created_at=created_at,
            text=data.get("body"),
            mentions=mentions,
            emoji_size=EmojiSize._from_tags(tags),
            sticker=sticker,
            attachments=attachments,
            unsent=unsent,
            forwarded=cls._get_forwarded_from_tags(tags),
        )


def graphql_to_extensible_attachment(data):
    story = data.get("story_attachment")
    if not story:
        return None

    target = story.get("target")
    if not target:
        return _attachment.UnsentMessage(id=data.get("legacy_attachment_id"))

    _type = target["__typename"]
    if _type == "MessageLocation":
        return _location.LocationAttachment._from_graphql(story)
    elif _type == "MessageLiveLocation":
        return _location.LiveLocationAttachment._from_graphql(story)
    elif _type in ["ExternalUrl", "Story"]:
        return _attachment.ShareAttachment._from_graphql(story)

    return None



import attr
import datetime
from ._common import attrs_event, Event, UnknownEvent, ThreadEvent
from .. import _util, _threads, _models

from typing import Sequence, Optional


@attrs_event
class ColorSet(ThreadEvent):
    """Somebody set the color in a thread."""

    #: The new color. Not limited to the ones in `ThreadABC.set_color`
    color: str
    #: When the color was set
    at: datetime.datetime

    @classmethod
    def _parse(cls, session, data):
        author, thread, at = cls._parse_metadata(session, data)
        color = _threads.ThreadABC._parse_color(data["untypedData"]["theme_color"])
        return cls(author=author, thread=thread, color=color, at=at)


@attrs_event
class EmojiSet(ThreadEvent):
    """Somebody set the emoji in a thread."""

    #: The new emoji
    emoji: str
    #: When the emoji was set
    at: datetime.datetime

    @classmethod
    def _parse(cls, session, data):
        author, thread, at = cls._parse_metadata(session, data)
        emoji = data["untypedData"]["thread_icon"]
        return cls(author=author, thread=thread, emoji=emoji, at=at)


@attrs_event
class NicknameSet(ThreadEvent):
    """Somebody set the nickname of a person in a thread."""

    #: The person whose nickname was set
    subject: str
    #: The new nickname. If ``None``, the nickname was cleared
    nickname: Optional[str]
    #: When the nickname was set
    at: datetime.datetime

    @classmethod
    def _parse(cls, session, data):
        author, thread, at = cls._parse_metadata(session, data)
        subject = _threads.User(
            session=session, id=data["untypedData"]["participant_id"]
        )
        nickname = data["untypedData"]["nickname"] or None  # None if ""
        return cls(
            author=author, thread=thread, subject=subject, nickname=nickname, at=at
        )


@attrs_event
class AdminsAdded(ThreadEvent):
    """Somebody added admins to a group."""

    #: The people that were set as admins
    added: Sequence["_threads.User"]
    #: When the admins were added
    at: datetime.datetime

    @classmethod
    def _parse(cls, session, data):
        author, thread, at = cls._parse_metadata(session, data)
        subject = _threads.User(session=session, id=data["untypedData"]["TARGET_ID"])
        return cls(author=author, thread=thread, added=[subject], at=at)


@attrs_event
class AdminsRemoved(ThreadEvent):
    """Somebody removed admins from a group."""

    #: The people that were removed as admins
    removed: Sequence["_threads.User"]
    #: When the admins were removed
    at: datetime.datetime

    @classmethod
    def _parse(cls, session, data):
        author, thread, at = cls._parse_metadata(session, data)
        subject = _threads.User(session=session, id=data["untypedData"]["TARGET_ID"])
        return cls(author=author, thread=thread, removed=[subject], at=at)


@attrs_event
class ApprovalModeSet(ThreadEvent):
    """Somebody changed the approval mode in a group."""

    require_admin_approval: bool
    #: When the approval mode was set
    at: datetime.datetime

    @classmethod
    def _parse(cls, session, data):
        author, thread, at = cls._parse_metadata(session, data)
        raa = data["untypedData"]["APPROVAL_MODE"] == "1"
        return cls(author=author, thread=thread, require_admin_approval=raa, at=at)


@attrs_event
class CallStarted(ThreadEvent):
    """Somebody started a call."""

    #: When the call was started
    at: datetime.datetime

    @classmethod
    def _parse(cls, session, data):
        author, thread, at = cls._parse_metadata(session, data)
        return cls(author=author, thread=thread, at=at)


@attrs_event
class CallEnded(ThreadEvent):
    """Somebody ended a call."""

    #: How long the call took
    duration: datetime.timedelta
    #: When the call ended
    at: datetime.datetime

    @classmethod
    def _parse(cls, session, data):
        author, thread, at = cls._parse_metadata(session, data)
        duration = _util.seconds_to_timedelta(int(data["untypedData"]["call_duration"]))
        return cls(author=author, thread=thread, duration=duration, at=at)


@attrs_event
class CallJoined(ThreadEvent):
    """Somebody joined a call."""

    #: When the call ended
    at: datetime.datetime

    @classmethod
    def _parse(cls, session, data):
        author, thread, at = cls._parse_metadata(session, data)
        return cls(author=author, thread=thread, at=at)


@attrs_event
class PollCreated(ThreadEvent):
    """Somebody created a group poll."""

    #: The new poll
    poll: "_models.Poll"
    #: When the poll was created
    at: datetime.datetime

    @classmethod
    def _parse(cls, session, data):
        author, thread, at = cls._parse_metadata(session, data)
        poll_data = _util.parse_json(data["untypedData"]["question_json"])
        poll = _models.Poll._from_graphql(session, poll_data)
        return cls(author=author, thread=thread, poll=poll, at=at)


@attrs_event
class PollVoted(ThreadEvent):
    """Somebody voted in a group poll."""

    #: The updated poll
    poll: "_models.Poll"
    #: Ids of the voted options
    added_ids: Sequence[str]
    #: Ids of the un-voted options
    removed_ids: Sequence[str]
    #: When the poll was voted in
    at: datetime.datetime

    @classmethod
    def _parse(cls, session, data):
        author, thread, at = cls._parse_metadata(session, data)
        poll_data = _util.parse_json(data["untypedData"]["question_json"])
        poll = _models.Poll._from_graphql(session, poll_data)
        added_ids = _util.parse_json(data["untypedData"]["added_option_ids"])
        removed_ids = _util.parse_json(data["untypedData"]["removed_option_ids"])
        return cls(
            author=author,
            thread=thread,
            poll=poll,
            added_ids=[str(x) for x in added_ids],
            removed_ids=[str(x) for x in removed_ids],
            at=at,
        )


@attrs_event
class PlanCreated(ThreadEvent):
    """Somebody created a plan in a group."""

    #: The new plan
    plan: "_models.PlanData"
    #: When the plan was created
    at: datetime.datetime

    @classmethod
    def _parse(cls, session, data):
        author, thread, at = cls._parse_metadata(session, data)
        plan = _models.PlanData._from_pull(session, data["untypedData"])
        return cls(author=author, thread=thread, plan=plan, at=at)


@attrs_event
class PlanEnded(ThreadEvent):
    """A plan ended."""

    #: The ended plan
    plan: "_models.PlanData"
    #: When the plan ended
    at: datetime.datetime

    @classmethod
    def _parse(cls, session, data):
        author, thread, at = cls._parse_metadata(session, data)
        plan = _models.PlanData._from_pull(session, data["untypedData"])
        return cls(author=author, thread=thread, plan=plan, at=at)


@attrs_event
class PlanEdited(ThreadEvent):
    """Somebody changed a plan in a group."""

    #: The updated plan
    plan: "_models.PlanData"
    #: When the plan was updated
    at: datetime.datetime

    @classmethod
    def _parse(cls, session, data):
        author, thread, at = cls._parse_metadata(session, data)
        plan = _models.PlanData._from_pull(session, data["untypedData"])
        return cls(author=author, thread=thread, plan=plan, at=at)


@attrs_event
class PlanDeleted(ThreadEvent):
    """Somebody removed a plan in a group."""

    #: The removed plan
    plan: "_models.PlanData"
    #: When the plan was removed
    at: datetime.datetime

    @classmethod
    def _parse(cls, session, data):
        author, thread, at = cls._parse_metadata(session, data)
        plan = _models.PlanData._from_pull(session, data["untypedData"])
        return cls(author=author, thread=thread, plan=plan, at=at)


@attrs_event
class PlanResponded(ThreadEvent):
    """Somebody responded to a plan in a group."""

    #: The plan that was responded to
    plan: "_models.PlanData"
    #: Whether the author will go to the plan or not
    take_part: bool
    #: When the plan was removed
    at: datetime.datetime

    @classmethod
    def _parse(cls, session, data):
        author, thread, at = cls._parse_metadata(session, data)
        plan = _models.PlanData._from_pull(session, data["untypedData"])
        take_part = data["untypedData"]["guest_status"] == "GOING"
        return cls(author=author, thread=thread, plan=plan, take_part=take_part, at=at)


def parse_admin_message(session, data):
    type_ = data["type"]
    if type_ == "change_thread_theme":
        return ColorSet._parse(session, data)
    elif type_ == "change_thread_icon":
        return EmojiSet._parse(session, data)
    elif type_ == "change_thread_nickname":
        return NicknameSet._parse(session, data)
    elif type_ == "change_thread_admins":
        event_type = data["untypedData"]["ADMIN_EVENT"]
        if event_type == "add_admin":
            return AdminsAdded._parse(session, data)
        elif event_type == "remove_admin":
            return AdminsRemoved._parse(session, data)
        else:
            pass
    elif type_ == "change_thread_approval_mode":
        return ApprovalModeSet._parse(session, data)
    elif type_ == "instant_game_update":
        pass  # TODO: This
    elif type_ == "messenger_call_log":  # Previously "rtc_call_log"
        event_type = data["untypedData"]["event"]
        if event_type == "group_call_started":
            return CallStarted._parse(session, data)
        elif event_type in ["group_call_ended", "one_on_one_call_ended"]:
            return CallEnded._parse(session, data)
        else:
            pass
    elif type_ == "participant_joined_group_call":
        return CallJoined._parse(session, data)
    elif type_ == "group_poll":
        event_type = data["untypedData"]["event_type"]
        if event_type == "question_creation":
            return PollCreated._parse(session, data)
        elif event_type == "update_vote":
            return PollVoted._parse(session, data)
        else:
            pass
    elif type_ == "lightweight_event_create":
        return PlanCreated._parse(session, data)
    elif type_ == "lightweight_event_notify":
        return PlanEnded._parse(session, data)
    elif type_ == "lightweight_event_update":
        return PlanEdited._parse(session, data)
    elif type_ == "lightweight_event_delete":
        return PlanDeleted._parse(session, data)
    elif type_ == "lightweight_event_rsvp":
        return PlanResponded._parse(session, data)
    return UnknownEvent(source="Delta type", data=data)



from ._common import *
from ._attachment import *
from ._file import *
from ._location import *
from ._plan import *
from ._poll import *
from ._quick_reply import *
from ._sticker import *
from ._message import *



import attr
import datetime
from ._abc import ThreadABC
from .._common import attrs_default
from .. import _session, _models

from typing import Optional


@attrs_default
class Page(ThreadABC):
    """Represents a Facebook page. Implements `ThreadABC`.

    Example:
        >>> page = fbchat.Page(session=session, id="1234")
    """

    # TODO: Implement pages properly, the implementation is lacking in a lot of places!

    #: The session to use when making requests.
    session: _session.Session
    #: The unique identifier of the page.
    id: str = attr.ib(converter=str)

    def _to_send_data(self):
        return {"other_user_fbid": self.id}

    def _copy(self) -> "Page":
        return Page(session=self.session, id=self.id)


@attrs_default
class PageData(Page):
    """Represents data about a Facebook page.

    Inherits `Page`, and implements `ThreadABC`.
    """

    #: The page's picture
    photo: _models.Image
    #: The name of the page
    name: str
    #: When the thread was last active / when the last message was sent
    last_active: Optional[datetime.datetime] = None
    #: Number of messages in the thread
    message_count: Optional[int] = None
    #: Set `Plan`
    plan: Optional[_models.PlanData] = None
    #: The page's custom URL
    url: Optional[str] = None
    #: The name of the page's location city
    city: Optional[str] = None
    #: Amount of likes the page has
    likes: Optional[int] = None
    #: Some extra information about the page
    sub_title: Optional[str] = None
    #: The page's category
    category: Optional[str] = None

    @classmethod
    def _from_graphql(cls, session, data):
        if data.get("profile_picture") is None:
            data["profile_picture"] = {}
        if data.get("city") is None:
            data["city"] = {}
        plan = None
        if data.get("event_reminders") and data["event_reminders"].get("nodes"):
            plan = _models.PlanData._from_graphql(
                session, data["event_reminders"]["nodes"][0]
            )

        return cls(
            session=session,
            id=data["id"],
            url=data.get("url"),
            city=data.get("city").get("name"),
            category=data.get("category_type"),
            photo=_models.Image._from_uri(data["profile_picture"]),
            name=data["name"],
            message_count=data.get("messages_count"),
            plan=plan,
        )



import attr
from . import Image, Attachment
from .._common import attrs_default

from typing import Optional


@attrs_default
class Sticker(Attachment):
    """Represents a Facebook sticker that has been sent to a thread as an attachment."""

    #: The sticker-pack's ID
    pack: Optional[str] = None
    #: Whether the sticker is animated
    is_animated: bool = False

    # If the sticker is animated, the following should be present
    #: URL to a medium spritemap
    medium_sprite_image: Optional[str] = None
    #: URL to a large spritemap
    large_sprite_image: Optional[str] = None
    #: The amount of frames present in the spritemap pr. row
    frames_per_row: Optional[int] = None
    #: The amount of frames present in the spritemap pr. column
    frames_per_col: Optional[int] = None
    #: The total amount of frames in the spritemap
    frame_count: Optional[int] = None
    #: The frame rate the spritemap is intended to be played in
    frame_rate: Optional[int] = None

    #: The sticker's image
    image: Optional[Image] = None
    #: The sticker's label/name
    label: Optional[str] = None

    @classmethod
    def _from_graphql(cls, data):
        if not data:
            return None

        return cls(
            id=data["id"],
            pack=data["pack"].get("id") if data.get("pack") else None,
            is_animated=bool(data.get("sprite_image")),
            medium_sprite_image=data["sprite_image"].get("uri")
            if data.get("sprite_image")
            else None,
            large_sprite_image=data["sprite_image_2x"].get("uri")
            if data.get("sprite_image_2x")
            else None,
            frames_per_row=data.get("frames_per_row"),
            frames_per_col=data.get("frames_per_column"),
            frame_count=data.get("frame_count"),
            frame_rate=data.get("frame_rate"),
            image=Image._from_url_or_none(data),
            label=data["label"] if data.get("label") else None,
        )



import json
import re
from ._common import log
from . import _util, _exception

# Shameless copy from https://stackoverflow.com/a/8730674
FLAGS = re.VERBOSE | re.MULTILINE | re.DOTALL
WHITESPACE = re.compile(r"[ \t\n\r]*", FLAGS)


class ConcatJSONDecoder(json.JSONDecoder):
    def decode(self, s, _w=WHITESPACE.match):
        s_len = len(s)

        objs = []
        end = 0
        while end != s_len:
            obj, end = self.raw_decode(s, idx=_w(s, end).end())
            end = _w(s, end).end()
            objs.append(obj)
        return objs


# End shameless copy


def queries_to_json(*queries):
    """
    Queries should be a list of GraphQL objects
    """
    rtn = {}
    for i, query in enumerate(queries):
        rtn["q{}".format(i)] = query
    return _util.json_minimal(rtn)


def response_to_json(text):
    text = _util.strip_json_cruft(text)  # Usually only needed in some error cases
    try:
        j = json.loads(text, cls=ConcatJSONDecoder)
    except Exception as e:
        raise _exception.ParseError("Error while parsing JSON", data=text) from e

    rtn = [None] * (len(j))
    for x in j:
        if "error_results" in x:
            del rtn[-1]
            continue
        _exception.handle_payload_error(x)
        [(key, value)] = x.items()
        _exception.handle_graphql_errors(value)
        if "response" in value:
            rtn[int(key[1:])] = value["response"]
        else:
            rtn[int(key[1:])] = value["data"]

    log.debug(rtn)

    return rtn


def from_query(query, params):
    return {"priority": 0, "q": query, "query_params": params}


def from_query_id(query_id, params):
    return {"query_id": query_id, "query_params": params}


def from_doc(doc, params):
    return {"doc": doc, "query_params": params}


def from_doc_id(doc_id, params):
    return {"doc_id": doc_id, "query_params": params}


FRAGMENT_USER = """
QueryFragment User: User {
    id,
    name,
    first_name,
    last_name,
    profile_picture.width(<pic_size>).height(<pic_size>) {
        uri
    },
    is_viewer_friend,
    url,
    gender,
    viewer_affinity
}
"""

FRAGMENT_GROUP = """
QueryFragment Group: MessageThread {
    name,
    thread_key {
        thread_fbid
    },
    image {
        uri
    },
    is_group_thread,
    all_participants {
        nodes {
            messaging_actor {
                __typename,
                id
            }
        }
    },
    customization_info {
        participant_customizations {
            participant_id,
            nickname
        },
        outgoing_bubble_color,
        emoji
    },
    thread_admins {
        id
    },
    group_approval_queue {
        nodes {
            requester {
                id
            }
        }
    },
    approval_mode,
    joinable_mode {
        mode,
        link
    },
    event_reminders {
        nodes {
            id,
            lightweight_event_creator {
                id
            },
            time,
            location_name,
            event_title,
            event_reminder_members {
                edges {
                    node {
                        id
                    },
                    guest_list_state
                }
            }
        }
    }
}
"""

FRAGMENT_PAGE = """
QueryFragment Page: Page {
    id,
    name,
    profile_picture.width(32).height(32) {
        uri
    },
    url,
    category_type,
    city {
        name
    }
}
"""

SEARCH_USER = (
    """
Query SearchUser(<search> = '', <limit> = 10) {
    entities_named(<search>) {
        search_results.of_type(user).first(<limit>) as users {
            nodes {
                @User
            }
        }
    }
}
"""
    + FRAGMENT_USER
)

SEARCH_GROUP = (
    """
Query SearchGroup(<search> = '', <limit> = 10, <pic_size> = 32) {
    viewer() {
        message_threads.with_thread_name(<search>).last(<limit>) as groups {
            nodes {
                @Group
            }
        }
    }
}
"""
    + FRAGMENT_GROUP
)

SEARCH_PAGE = (
    """
Query SearchPage(<search> = '', <limit> = 10) {
    entities_named(<search>) {
        search_results.of_type(page).first(<limit>) as pages {
            nodes {
                @Page
            }
        }
    }
}
"""
    + FRAGMENT_PAGE
)

SEARCH_THREAD = (
    """
Query SearchThread(<search> = '', <limit> = 10) {
    entities_named(<search>) {
        search_results.first(<limit>) as threads {
            nodes {
                __typename,
                @User,
                @Group,
                @Page
            }
        }
    }
}
"""
    + FRAGMENT_USER
    + FRAGMENT_GROUP
    + FRAGMENT_PAGE
)



import attr
import aiohttp
import asyncio

from typing import Any, Optional

# Not frozen, since that doesn't work in PyPy
@attr.s(slots=True, auto_exc=True, auto_attribs=True)
class FacebookError(Exception):
    """Base class for all custom exceptions raised by ``fbchat``.

    All exceptions in the module inherit this.
    """

    #: A message describing the error
    message: str


@attr.s(slots=True, auto_exc=True, auto_attribs=True)
class HTTPError(FacebookError):
    """Base class for errors with the HTTP(s) connection to Facebook."""

    #: The returned HTTP status code, if relevant
    status_code: Optional[int] = None

    def __str__(self):
        if not self.status_code:
            return self.message
        return "Got {} response: {}".format(self.status_code, self.message)


@attr.s(slots=True, auto_exc=True, auto_attribs=True)
class ParseError(FacebookError):
    """Raised when we fail parsing a response from Facebook.

    This may contain sensitive data, so should not be logged to file.
    """

    data: Any
    """The data that triggered the error.

    The format of this cannot be relied on, it's only for debugging purposes.
    """

    def __str__(self):
        msg = "{}. Please report this, along with the data below!\n{}"
        return msg.format(self.message, self.data)


@attr.s(slots=True, auto_exc=True, auto_attribs=True)
class NotLoggedIn(FacebookError):
    """Raised by Facebook if the client has been logged out."""


@attr.s(slots=True, auto_exc=True, auto_attribs=True)
class NotConnected(FacebookError):
    """Raised by Facebook if the client has been logged out."""


@attr.s(slots=True, auto_exc=True, auto_attribs=True)
class ExternalError(FacebookError):
    """Base class for errors that Facebook return."""

    #: The error message that Facebook returned (Possibly in the user's own language)
    description: str
    #: The error code that Facebook returned
    code: Optional[int] = None

    def __str__(self):
        if self.code:
            return "#{} {}: {}".format(self.code, self.message, self.description)
        return "{}: {}".format(self.message, self.description)


@attr.s(slots=True, auto_exc=True, auto_attribs=True)
class GraphQLError(ExternalError):
    """Raised by Facebook if there was an error in the GraphQL query."""

    # TODO: Handle multiple errors

    #: Query debug information
    debug_info: Optional[str] = None

    def __str__(self):
        if self.debug_info:
            return "{}, {}".format(super().__str__(), self.debug_info)
        return super().__str__()


@attr.s(slots=True, auto_exc=True, auto_attribs=True)
class InvalidParameters(ExternalError):
    """Raised by Facebook if:

    - Some function supplied invalid parameters.
    - Some content is not found.
    - Some content is no longer available.
    """


@attr.s(slots=True, auto_exc=True, auto_attribs=True)
class PleaseRefresh(ExternalError):
    """Raised by Facebook if the client has been inactive for too long.

    This error usually happens after 1-2 days of inactivity.
    """

    code: int = 1357004


def handle_payload_error(j):
    if "error" not in j:
        return
    code = j["error"]
    if code == 1357001:
        raise NotLoggedIn(j["errorSummary"])
    elif code == 1357004:
        error_cls = PleaseRefresh
    elif code in (1357031, 1545010, 1545003):
        error_cls = InvalidParameters
    else:
        error_cls = ExternalError
    raise error_cls(j["errorSummary"], description=j["errorDescription"], code=code)


def handle_graphql_errors(j):
    errors = []
    if j.get("error"):
        errors = [j["error"]]
    if "errors" in j:
        errors = j["errors"]
    if errors:
        error = errors[0]  # TODO: Handle multiple errors
        # TODO: Use `severity` and `description`
        raise GraphQLError(
            # TODO: What data is always available?
            message=error.get("summary", "Unknown error"),
            description=error.get("message", ""),
            code=error.get("code"),
            debug_info=error.get("debug_info"),
        )


def handle_http_error(code):
    if code == 404:
        raise HTTPError(
            "This might be because you provided an invalid id"
            + " (Facebook usually require integer ids)",
            status_code=code,
        )
    if code == 500:
        raise HTTPError(
            "There is probably an error on the endpoint, or it might be rate limited",
            status_code=code,
        )
    if 400 <= code < 600:
        raise HTTPError("Failed sending request", status_code=code)


def handle_requests_error(e):
    if isinstance(e, (aiohttp.ClientConnectionError, aiohttp.ServerConnectionError)):
        raise HTTPError("Connection error") from e
    if isinstance(e, aiohttp.ClientResponseError):
        pass  # Raised when using .raise_for_status, so should never happen
    if isinstance(e, aiohttp.InvalidURL):
        pass  # Should never happen, we always prove valid URLs
    if isinstance(e, aiohttp.TooManyRedirects):
        pass  # TODO: Consider using allow_redirects=False to prevent this
    if isinstance(e, (aiohttp.ServerTimeoutError, asyncio.TimeoutError)):
        pass  # Should never happen, we don't set timeouts

    raise HTTPError("Requests error") from e



import attr
import datetime

from ._common import log, attrs_default
from . import _exception, _util, _graphql, _session, _threads, _models

from typing import Sequence, Iterable, Tuple, Optional, Set, BinaryIO, AsyncIterator


@attrs_default
class Client:
    """A client for Facebook Messenger.

    This contains methods that are generally needed to interact with Facebook.

    Example:
        Create a new client instance.

        >>> client = fbchat.Client(session=session)
    """

    #: The session to use when making requests.
    session: _session.Session

    async def fetch_users(self) -> Sequence[_threads.UserData]:
        """Fetch users the client is currently chatting with.

        This is very close to your friend list, with the follow differences:

        It differs by including users that you're not friends with, but have chatted
        with before, and by including accounts that are "Messenger Only".

        But does not include deactivated, deleted or memorialized users (logically,
        since you can't chat with those).

        The order these are returned is arbitrary.

        Example:
            Get the name of an arbitrary user that you're currently chatting with.

            >>> users = client.fetch_users()
            >>> users[0].name
            "A user"
        """
        data = {"viewer": self.session.user.id}
        j = await self.session._payload_post("/chat/user_info_all", data)

        users = []
        for data in j.values():
            if data["type"] not in ["user", "friend"] or data["id"] in ["0", 0]:
                log.warning("Invalid user data %s", data)
                continue  # Skip invalid users
            users.append(_threads.UserData._from_all_fetch(self.session, data))
        return users

    async def search_for_users(self, name: str, limit: int) -> Iterable[_threads.UserData]:
        """Find and get users by their name.

        The returned users are ordered by relevance.

        Args:
            name: Name of the user
            limit: The max. amount of users to fetch

        Example:
            Get the full name of the first found user.

            >>> (user,) = client.search_for_users("user", limit=1)
            >>> user.name
            "A user"
        """
        params = {"search": name, "limit": limit}
        (j,) = await self.session._graphql_requests(
            _graphql.from_query(_graphql.SEARCH_USER, params)
        )

        return (
            _threads.UserData._from_graphql(self.session, node)
            for node in j[name]["users"]["nodes"]
        )

    async def search_for_pages(self, name: str, limit: int) -> Iterable[_threads.PageData]:
        """Find and get pages by their name.

        The returned pages are ordered by relevance.

        Args:
            name: Name of the page
            limit: The max. amount of pages to fetch

        Example:
            Get the full name of the first found page.

            >>> (page,) = client.search_for_pages("page", limit=1)
            >>> page.name
            "A page"
        """
        params = {"search": name, "limit": limit}
        (j,) = await self.session._graphql_requests(
            _graphql.from_query(_graphql.SEARCH_PAGE, params)
        )

        return (
            _threads.PageData._from_graphql(self.session, node)
            for node in j[name]["pages"]["nodes"]
        )

    async def search_for_groups(self, name: str, limit: int) -> Iterable[_threads.GroupData]:
        """Find and get group threads by their name.

        The returned groups are ordered by relevance.

        Args:
            name: Name of the group thread
            limit: The max. amount of groups to fetch

        Example:
            Get the full name of the first found group.

            >>> (group,) = client.search_for_groups("group", limit=1)
            >>> group.name
            "A group"
        """
        params = {"search": name, "limit": limit}
        (j,) = await self.session._graphql_requests(
            _graphql.from_query(_graphql.SEARCH_GROUP, params)
        )

        return (
            _threads.GroupData._from_graphql(self.session, node)
            for node in j["viewer"]["groups"]["nodes"]
        )

    async def search_for_threads(self, name: str, limit: int) -> Iterable[_threads.ThreadABC]:
        """Find and get threads by their name.

        The returned threads are ordered by relevance.

        Args:
            name: Name of the thread
            limit: The max. amount of threads to fetch

        Example:
            Search for a user, and get the full name of the first found result.

            >>> (user,) = client.search_for_threads("user", limit=1)
            >>> assert isinstance(user, fbchat.User)
            >>> user.name
            "A user"
        """
        params = {"search": name, "limit": limit}
        (j,) = await self.session._graphql_requests(
            _graphql.from_query(_graphql.SEARCH_THREAD, params)
        )

        for node in j[name]["threads"]["nodes"]:
            if node["__typename"] == "User":
                yield _threads.UserData._from_graphql(self.session, node)
            elif node["__typename"] == "MessageThread":
                # MessageThread => Group thread
                yield _threads.GroupData._from_graphql(self.session, node)
            elif node["__typename"] == "Page":
                yield _threads.PageData._from_graphql(self.session, node)
            elif node["__typename"] == "Group":
                # We don't handle Facebook "Groups"
                pass
            else:
                log.warning(
                    "Unknown type {} in {}".format(repr(node["__typename"]), node)
                )

    async def _search_messages(self, query, offset, limit):
        data = {"query": query, "offset": offset, "limit": limit}
        j = await self.session._payload_post("/ajax/mercury/search_snippets.php?dpr=1", data)

        total_snippets = j["search_snippets"][query]

        rtn = []
        for node in j["graphql_payload"]["message_threads"]:
            type_ = node["thread_type"]
            if type_ == "GROUP":
                thread = _threads.Group(
                    session=self.session, id=node["thread_key"]["thread_fbid"]
                )
            elif type_ == "ONE_TO_ONE":
                thread = _threads.Thread(
                    session=self.session, id=node["thread_key"]["other_user_id"]
                )
                # if True:  # TODO: This check!
                #     thread = _threads.UserData._from_graphql(self.session, node)
                # else:
                #     thread = _threads.PageData._from_graphql(self.session, node)
            else:
                thread = None
                log.warning("Unknown thread type %s, data: %s", type_, node)

            if thread:
                rtn.append((thread, total_snippets[thread.id]["num_total_snippets"]))
            else:
                rtn.append((None, 0))

        return rtn

    async def search_messages(
        self, query: str, limit: Optional[int]
    ) -> AsyncIterator[Tuple[_threads.ThreadABC, int]]:
        """Search for messages in all threads.

        Intended to be used alongside `ThreadABC.search_messages`.

        Warning! If someone send a message to a thread that matches the query, while
        we're searching, some snippets will get returned twice, and some will be lost.

        This is fundamentally not fixable, it's just how the endpoint is implemented.

        Args:
            query: Text to search for
            limit: Max. number of items to retrieve. If ``None``, all will be retrieved

        Example:
            Search for messages, and print the amount of snippets in each thread.

            >>> for thread, count in client.search_messages("abc", limit=3):
            ...     print(f"{thread.id} matched the search {count} time(s)")
            ...
            1234 matched the search 2 time(s)
            2345 matched the search 1 time(s)
            3456 matched the search 100 time(s)

        Return:
            Iterable with tuples of threads, and the total amount of matches.
        """
        offset = 0
        # The max limit is measured empirically to ~500, safe default chosen below
        for limit in _util.get_limits(limit, max_limit=100):
            data = await self._search_messages(query, offset, limit)
            for thread, total_snippets in data:
                if thread:
                    yield (thread, total_snippets)
            if len(data) < limit:
                return  # No more data to fetch
            offset += limit

    async def _fetch_info(self, *ids):
        data = {"ids[{}]".format(i): _id for i, _id in enumerate(ids)}
        j = await self.session._payload_post("/chat/user_info/", data)

        if j.get("profiles") is None:
            raise _exception.ParseError("No users/pages returned", data=j)

        entries = {}
        for _id in j["profiles"]:
            k = j["profiles"][_id]
            if k["type"] in ["user", "friend"]:
                entries[_id] = {
                    "id": _id,
                    "url": k.get("uri"),
                    "first_name": k.get("firstName"),
                    "is_viewer_friend": k.get("is_friend"),
                    "gender": k.get("gender"),
                    "profile_picture": {"uri": k.get("thumbSrc")},
                    "name": k.get("name"),
                }
            elif k["type"] == "page":
                entries[_id] = {
                    "id": _id,
                    "url": k.get("uri"),
                    "profile_picture": {"uri": k.get("thumbSrc")},
                    "name": k.get("name"),
                }
            else:
                raise _exception.ParseError("Unknown thread type", data=k)

        log.debug(entries)
        return entries

    async def fetch_thread_info(self, ids: Iterable[str]) -> AsyncIterator[_threads.ThreadABC]:
        """Fetch threads' info from IDs, unordered.

        Warning:
            Sends two requests if users or pages are present, to fetch all available info!

        Args:
            ids: Thread ids to query

        Example:
            Get data about the user with id "4".

            >>> (user,) = client.fetch_thread_info(["4"])
            >>> user.name
            "Mark Zuckerberg"
        """
        ids = list(ids)
        queries = []
        for thread_id in ids:
            params = {
                "id": thread_id,
                "message_limit": 0,
                "load_messages": False,
                "load_read_receipts": False,
                "before": None,
            }
            queries.append(_graphql.from_doc_id("2147762685294928", params))

        j = await self.session._graphql_requests(*queries)

        for i, entry in enumerate(j):
            if entry.get("message_thread") is None:
                # If you don't have an existing thread with this person, attempt to retrieve user data anyways
                j[i]["message_thread"] = {
                    "thread_key": {"other_user_id": ids[i]},
                    "thread_type": "ONE_TO_ONE",
                }

        pages_and_user_ids = [
            k["message_thread"]["thread_key"]["other_user_id"]
            for k in j
            if k["message_thread"].get("thread_type") == "ONE_TO_ONE"
        ]
        pages_and_users = {}
        if len(pages_and_user_ids) != 0:
            pages_and_users = await self._fetch_info(*pages_and_user_ids)

        for i, entry in enumerate(j):
            entry = entry["message_thread"]
            if entry.get("thread_type") == "GROUP":
                _id = entry["thread_key"]["thread_fbid"]
                yield _threads.GroupData._from_graphql(self.session, entry)
            elif entry.get("thread_type") == "ONE_TO_ONE":
                _id = entry["thread_key"]["other_user_id"]
                if pages_and_users.get(_id) is None:
                    raise _exception.ParseError(
                        "Could not fetch thread {}".format(_id), data=pages_and_users
                    )
                entry.update(pages_and_users[_id])
                if "first_name" in entry:
                    yield _threads.UserData._from_graphql(self.session, entry)
                else:
                    yield _threads.PageData._from_graphql(self.session, entry)
            else:
                raise _exception.ParseError("Unknown thread type", data=entry)

    async def _fetch_threads(self, limit, before, folders):
        params = {
            "limit": limit,
            "tags": folders,
            "before": _util.datetime_to_millis(before) if before else None,
            "includeDeliveryReceipts": True,
            "includeSeqID": False,
        }
        (j,) = await self.session._graphql_requests(
            _graphql.from_doc_id("1349387578499440", params)
        )

        rtn = []
        for node in j["viewer"]["message_threads"]["nodes"]:
            _type = node.get("thread_type")
            if _type == "GROUP":
                rtn.append(_threads.GroupData._from_graphql(self.session, node))
            elif _type == "ONE_TO_ONE":
                rtn.append(_threads.UserData._from_thread_fetch(self.session, node))
            else:
                rtn.append(None)
                log.warning("Unknown thread type: %s, data: %s", _type, node)
        return rtn

    async def fetch_threads(
        self,
        limit: Optional[int],
        location: _models.ThreadLocation = _models.ThreadLocation.INBOX,
    ) -> AsyncIterator[_threads.ThreadABC]:
        """Fetch the client's thread list.

        The returned threads are ordered by last active first.

        Args:
            limit: Max. number of threads to retrieve. If ``None``, all threads will be
                retrieved.
            location: INBOX, PENDING, ARCHIVED or OTHER

        Example:
            Fetch the last three threads that the user chatted with.

            >>> for thread in client.fetch_threads(limit=3):
            ...     print(f"{thread.id}: {thread.name}")
            ...
            1234: A user
            2345: A group
            3456: A page
        """
        # This is measured empirically as 837, safe default chosen below
        MAX_BATCH_LIMIT = 100

        # TODO: Clean this up after implementing support for more threads types
        seen_ids = set()  # type: Set[str]
        before = None
        for limit in _util.get_limits(limit, MAX_BATCH_LIMIT):
            threads = await self._fetch_threads(limit, before, [location.value])

            before = None
            for thread in threads:
                # Don't return seen and unknown threads
                if thread and thread.id not in seen_ids:
                    seen_ids.add(thread.id)
                    # TODO: Ensure type-wise that .last_active is available
                    before = thread.last_active
                    yield thread

            if len(threads) < MAX_BATCH_LIMIT:
                return  # No more data to fetch

            # We check this here in case _fetch_threads only returned `None` threads
            if not before:
                raise ValueError("Too many unknown threads.")

    async def fetch_unread(self) -> Sequence[_threads.ThreadABC]:
        """Fetch unread threads.

        Warning:
            This is not finished, and the API may change at any point!
        """
        at = datetime.datetime.utcnow()
        form = {
            "folders[0]": "inbox",
            "client": "mercury",
            "last_action_timestamp": _util.datetime_to_millis(at),
            # 'last_action_timestamp': 0
        }
        j = await self.session._payload_post("/ajax/mercury/unread_threads.php", form)

        result = j["unread_thread_fbids"][0]
        # TODO: Parse Pages?
        return [
            _threads.Group(session=self.session, id=id_)
            for id_ in result["thread_fbids"]
        ] + [
            _threads.User(session=self.session, id=id_)
            for id_ in result["other_user_fbids"]
        ]

    async def fetch_unseen(self) -> Sequence[_threads.ThreadABC]:
        """Fetch unseen / new threads.

        Warning:
            This is not finished, and the API may change at any point!
        """
        j = await self.session._payload_post("/mercury/unseen_thread_ids/", {})

        result = j["unseen_thread_fbids"][0]
        # TODO: Parse Pages?
        return [
            _threads.Group(session=self.session, id=id_)
            for id_ in result["thread_fbids"]
        ] + [
            _threads.User(session=self.session, id=id_)
            for id_ in result["other_user_fbids"]
        ]

    async def fetch_image_url(self, image_id: str) -> str:
        """Fetch URL to download the original image from an image attachment ID.

        Args:
            image_id: The image you want to fetch

        Example:
            >>> client.fetch_image_url("1234")
            "https://scontent-arn1-1.xx.fbcdn.net/v/t1.123-4/1_23_45_n.png?..."

        Returns:
            An URL where you can download the original image
        """
        image_id = str(image_id)
        data = {"photo_id": str(image_id)}
        j = await self.session._post("/mercury/attachments/photo/", data)
        _exception.handle_payload_error(j)

        if "jsmods" not in j:
            raise _exception.ParseError("No jsmods when fetching image URL", data=j)
        require = _util.get_jsmods_require(j["jsmods"]["require"])
        if "ServerRedirect.redirectPageTo" not in require:
            raise _exception.ParseError("Could not fetch image URL", data=j)
        # Return the first argument
        return require["ServerRedirect.redirectPageTo"][0]

    async def _get_private_data(self):
        (j,) = await self.session._graphql_requests(
            _graphql.from_doc_id("1868889766468115", {})
        )
        return j["viewer"]

    async def get_phone_numbers(self) -> Sequence[str]:
        """Fetch the user's phone numbers."""
        data = await self._get_private_data()
        return [
            j["phone_number"]["universal_number"] for j in data["user"]["all_phones"]
        ]

    async def get_emails(self) -> Sequence[str]:
        """Fetch the user's emails."""
        data = await self._get_private_data()
        return [j["display_email"] for j in data["all_emails"]]

    async def upload(
        self, files: Iterable[Tuple[str, BinaryIO, str]], voice_clip: bool = False
    ) -> Sequence[Tuple[str, str]]:
        """Upload files to Facebook.

        `files` should be a list of files that requests can upload, see
        `requests.request <https://docs.python-requests.org/en/master/api/#requests.request>`_.

        Example:
            >>> with open("file.txt", "rb") as f:
            ...     (file,) = client.upload([("file.txt", f, "text/plain")])
            ...
            >>> file
            ("1234", "text/plain")
        Return:
            Tuples with a file's ID and mimetype.
            This result can be passed straight on to `ThreadABC.send_files`, or used in
            `Group.set_image`.
        """
        file_dict = {"upload_{}".format(i): f for i, f in enumerate(files)}

        data = {"voice_clip": voice_clip}

        j = await self.session._payload_post(
            "https://upload.messenger.com/ajax/mercury/upload.php",
            data,
            files=file_dict,
        )

        if len(j["metadata"]) != len(file_dict):
            raise _exception.ParseError("Some files could not be uploaded", data=j)

        return [
            (str(item[_util.mimetype_to_key(item["filetype"])]), item["filetype"])
            for item in j["metadata"]
        ]

    async def mark_as_delivered(self, message: _models.Message):
        """Mark a message as delivered.

        Warning:
            This is not finished, and the API may change at any point!

        Args:
            message: The message to set as delivered
        """
        data = {
            "message_ids[0]": message.id,
            "thread_ids[%s][0]" % message.thread.id: message.id,
        }
        j = await self.session._payload_post("/ajax/mercury/delivery_receipts.php", data)

    async def _read_status(self, read, threads, at):
        data = {
            "watermarkTimestamp": _util.datetime_to_millis(at),
            "shouldSendReadReceipt": "true",
        }

        for thread in threads:
            data["ids[{}]".format(thread.id)] = "true" if read else "false"

        j = await self.session._payload_post("/ajax/mercury/change_read_status.php", data)

    async def mark_as_read(
        self, threads: Iterable[_threads.ThreadABC], at: datetime.datetime
    ):
        """Mark threads as read.

        All messages inside the specified threads will be marked as read.

        Args:
            threads: Threads to set as read
            at: Timestamp to signal the read cursor at
        """
        return await self._read_status(True, threads, at)

    async def mark_as_unread(
        self, threads: Iterable[_threads.ThreadABC], at: datetime.datetime
    ):
        """Mark threads as unread.

        All messages inside the specified threads will be marked as unread.

        Args:
            threads: Threads to set as unread
            at: Timestamp to signal the read cursor at
        """
        return await self._read_status(False, threads, at)

    async def mark_as_seen(self, at: datetime.datetime):
        # TODO: Documenting this
        data = {"seen_timestamp": _util.datetime_to_millis(at)}
        j = await self.session._payload_post("/ajax/mercury/mark_seen.php", data)

    async def move_threads(
        self, location: _models.ThreadLocation, threads: Iterable[_threads.ThreadABC]
    ):
        """Move threads to specified location.

        Args:
            location: INBOX, PENDING, ARCHIVED or OTHER
            threads: Threads to move
        """
        if location == _models.ThreadLocation.PENDING:
            location = _models.ThreadLocation.OTHER

        if location == _models.ThreadLocation.ARCHIVED:
            data_archive = {}
            data_unpin = {}
            for thread in threads:
                data_archive["ids[{}]".format(thread.id)] = "true"
                data_unpin["ids[{}]".format(thread.id)] = "false"
            j_archive = await self.session._payload_post(
                "/ajax/mercury/change_archived_status.php?dpr=1", data_archive
            )
            j_unpin = await self.session._payload_post(
                "/ajax/mercury/change_pinned_status.php?dpr=1", data_unpin
            )
        else:
            data = {}
            for i, thread in enumerate(threads):
                data["{}[{}]".format(location.name.lower(), i)] = thread.id
            j = await self.session._payload_post("/ajax/mercury/move_threads.php", data)

    async def delete_threads(self, threads: Iterable[_threads.ThreadABC]):
        """Bulk delete threads.

        Args:
            threads: Threads to delete

        Example:
            >>> group = fbchat.Group(session=session, id="1234")
            >>> client.delete_threads([group])
        """
        await _threads.ThreadABC._delete_many(self.session, (t.id for t in threads))

    async def delete_messages(self, messages: Iterable[_models.Message]):
        """Bulk delete specified messages.

        Args:
            messages: Messages to delete

        Example:
            >>> message1 = fbchat.Message(thread=thread, id="1234")
            >>> message2 = fbchat.Message(thread=thread, id="2345")
            >>> client.delete_threads([message1, message2])
        """
        await _models.Message._delete_many(self.session, (m.id for m in messages))



import abc
import attr
import collections
import datetime
from .._common import log, attrs_default
from .. import _util, _exception, _session, _graphql, _models
from typing import MutableMapping, Mapping, Any, Iterable, Tuple, Optional, AsyncGenerator


DEFAULT_COLOR = "#0084ff"
SETABLE_COLORS = (
    DEFAULT_COLOR,
    "#44bec7",
    "#ffc300",
    "#fa3c4c",
    "#d696bb",
    "#6699cc",
    "#13cf13",
    "#ff7e29",
    "#e68585",
    "#7646ff",
    "#20cef5",
    "#67b868",
    "#d4a88c",
    "#ff5ca1",
    "#a695c7",
    "#ff7ca8",
    "#1adb5b",
    "#f01d6a",
    "#ff9c19",
    "#0edcde",
)


class ThreadABC(metaclass=abc.ABCMeta):
    """Implemented by thread-like classes.

    This is private to implement.
    """

    @property
    @abc.abstractmethod
    def session(self) -> _session.Session:
        """The session to use when making requests."""
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def id(self) -> str:
        """The unique identifier of the thread."""
        raise NotImplementedError

    @abc.abstractmethod
    def _to_send_data(self) -> MutableMapping[str, str]:
        raise NotImplementedError

    # Note:
    # You can go out of Facebook's spec with `self.session._do_send_request`!
    #
    # A few examples:
    # - You can send a sticker and an emoji at the same time
    # - You can wave, send a sticker and text at the same time
    # - You can reply to a message with a sticker
    #
    # We won't support those use cases, it'll make for a confusing API!
    # If we absolutely need to in the future, we can always add extra functionality

    @abc.abstractmethod
    def _copy(self) -> "ThreadABC":
        """It may or may not be a good idea to attach the current thread to new objects.

        So for now, we use this method to create a new thread.

        This should return the minimal representation of the thread (e.g. not UserData).
        """
        raise NotImplementedError

    async def fetch(self):
        # TODO: This
        raise NotImplementedError

    async def wave(self, first: bool = True) -> str:
        """Wave hello to the thread.

        Args:
            first: Whether to wave first or wave back

        Example:
            Wave back to the thread.

            >>> thread.wave(False)
        """
        data = self._to_send_data()
        data["action_type"] = "ma-type:user-generated-message"
        data["lightweight_action_attachment[lwa_state]"] = (
            "INITIATED" if first else "RECIPROCATED"
        )
        data["lightweight_action_attachment[lwa_type]"] = "WAVE"
        message_id, thread_id = await self.session._do_send_request(data)
        return message_id

    async def send_text(
        self,
        text: str,
        mentions: Iterable["_models.Mention"] = None,
        files: Iterable[Tuple[str, str]] = None,
        reply_to_id: str = None,
    ) -> str:
        """Send a message to the thread.

        Args:
            text: Text to send
            mentions: Optional mentions
            files: Optional tuples, each containing an uploaded file's ID and mimetype.
                See `ThreadABC.send_files` for an example.
            reply_to_id: Optional message to reply to

        Example:
            >>> mention = fbchat.Mention(thread_id="1234", offset=5, length=2)
            >>> thread.send_text("A message", mentions=[mention])

        Returns:
            The sent message
        """
        data = self._to_send_data()
        data["action_type"] = "ma-type:user-generated-message"
        if text is not None:  # To support `send_files`
            data["body"] = text

        for i, mention in enumerate(mentions or ()):
            data.update(mention._to_send_data(i))

        if files:
            data["has_attachment"] = True

        for i, (file_id, mimetype) in enumerate(files or ()):
            data["{}s[{}]".format(_util.mimetype_to_key(mimetype), i)] = file_id

        if reply_to_id:
            data["replied_to_message_id"] = reply_to_id

        return await self.session._do_send_request(data)

    async def send_emoji(self, emoji: str, size: "_models.EmojiSize") -> str:
        """Send an emoji to the thread.

        Args:
            emoji: The emoji to send
            size: The size of the emoji

        Example:
            >>> thread.send_emoji("😀", size=fbchat.EmojiSize.LARGE)

        Returns:
            The sent message
        """
        data = self._to_send_data()
        data["action_type"] = "ma-type:user-generated-message"
        data["body"] = emoji
        data["tags[0]"] = "hot_emoji_size:{}".format(size.name.lower())
        return await self.session._do_send_request(data)

    async def send_sticker(self, sticker_id: str) -> str:
        """Send a sticker to the thread.

        Args:
            sticker_id: ID of the sticker to send

        Example:
            Send a sticker with the id "1889713947839631"

            >>> thread.send_sticker("1889713947839631")

        Returns:
            The sent message
        """
        data = self._to_send_data()
        data["action_type"] = "ma-type:user-generated-message"
        data["sticker_id"] = sticker_id
        return await self.session._do_send_request(data)

    async def _send_location(self, current, latitude, longitude):
        data = self._to_send_data()
        data["action_type"] = "ma-type:user-generated-message"
        data["location_attachment[coordinates][latitude]"] = latitude
        data["location_attachment[coordinates][longitude]"] = longitude
        data["location_attachment[is_current_location]"] = current
        return await self.session._do_send_request(data)

    async def send_location(self, latitude: float, longitude: float):
        """Send a given location to a thread as the user's current location.

        Args:
            latitude: The location latitude
            longitude: The location longitude

        Example:
            Send a location in London, United Kingdom.

            >>> thread.send_location(51.5287718, -0.2416815)
        """
        await self._send_location(True, latitude=latitude, longitude=longitude)

    async def send_pinned_location(self, latitude: float, longitude: float):
        """Send a given location to a thread as a pinned location.

        Args:
            latitude: The location latitude
            longitude: The location longitude

        Example:
            Send a pinned location in Beijing, China.

            >>> thread.send_pinned_location(39.9390731, 116.117273)
        """
        await self._send_location(False, latitude=latitude, longitude=longitude)

    async def send_files(self, files: Iterable[Tuple[str, str]]):
        """Send files from file IDs to a thread.

        `files` should be a list of tuples, with a file's ID and mimetype.

        Example:
            Upload and send a video to a thread.

            >>> with open("video.mp4", "rb") as f:
            ...     files = client.upload([("video.mp4", f, "video/mp4")])
            >>>
            >>> thread.send_files(files)
        """
        return await self.send_text(text=None, files=files)

    # xmd = {"quick_replies": []}
    # for quick_reply in quick_replies:
    #     # TODO: Move this to `_quick_reply.py`
    #     q = dict()
    #     q["content_type"] = quick_reply._type
    #     q["payload"] = quick_reply.payload
    #     q["external_payload"] = quick_reply.external_payload
    #     q["data"] = quick_reply.data
    #     if quick_reply.is_response:
    #         q["ignore_for_webhook"] = False
    #     if isinstance(quick_reply, _quick_reply.QuickReplyText):
    #         q["title"] = quick_reply.title
    #     if not isinstance(quick_reply, _quick_reply.QuickReplyLocation):
    #         q["image_url"] = quick_reply.image_url
    #     xmd["quick_replies"].append(q)
    # if len(quick_replies) == 1 and quick_replies[0].is_response:
    #     xmd["quick_replies"] = xmd["quick_replies"][0]
    # data["platform_xmd"] = _util.json_minimal(xmd)

    # TODO: This!
    # def quick_reply(self, quick_reply: QuickReply, payload=None):
    #     """Reply to chosen quick reply.
    #
    #     Args:
    #         quick_reply: Quick reply to reply to
    #         payload: Optional answer to the quick reply
    #     """
    #     if isinstance(quick_reply, QuickReplyText):
    #         new = QuickReplyText(
    #             payload=quick_reply.payload,
    #             external_payload=quick_reply.external_payload,
    #             data=quick_reply.data,
    #             is_response=True,
    #             title=quick_reply.title,
    #             image_url=quick_reply.image_url,
    #         )
    #         return self.send(Message(text=quick_reply.title, quick_replies=[new]))
    #     elif isinstance(quick_reply, QuickReplyLocation):
    #         if not isinstance(payload, LocationAttachment):
    #             raise TypeError("Payload must be an instance of `LocationAttachment`")
    #         return self.send_location(payload)
    #     elif isinstance(quick_reply, QuickReplyEmail):
    #         new = QuickReplyEmail(
    #             payload=payload if payload else self.get_emails()[0],
    #             external_payload=quick_reply.payload,
    #             data=quick_reply.data,
    #             is_response=True,
    #             image_url=quick_reply.image_url,
    #         )
    #         return self.send(Message(text=payload, quick_replies=[new]))
    #     elif isinstance(quick_reply, QuickReplyPhoneNumber):
    #         new = QuickReplyPhoneNumber(
    #             payload=payload if payload else self.get_phone_numbers()[0],
    #             external_payload=quick_reply.payload,
    #             data=quick_reply.data,
    #             is_response=True,
    #             image_url=quick_reply.image_url,
    #         )
    #         return self.send(Message(text=payload, quick_replies=[new]))

    async def _search_messages(self, query, offset, limit):
        data = {
            "query": query,
            "snippetOffset": offset,
            "snippetLimit": limit,
            "identifier": "thread_fbid",
            "thread_fbid": self.id,
        }
        j = await self.session._payload_post("/ajax/mercury/search_snippets.php?dpr=1", data)

        result = j["search_snippets"][query].get(self.id)
        if not result:
            return (0, [])

        thread = self._copy()
        snippets = [
            _models.MessageSnippet._parse(thread, snippet)
            for snippet in result["snippets"]
        ]
        return (result["num_total_snippets"], snippets)

    async def search_messages(
        self, query: str, limit: int
    ) -> Iterable[_models.MessageSnippet]:
        """Find and get message IDs by query.

        Warning! If someone send a message to the thread that matches the query, while
        we're searching, some snippets will get returned twice.

        This is fundamentally not fixable, it's just how the endpoint is implemented.

        The returned message snippets are ordered by last sent first.

        Args:
            query: Text to search for
            limit: Max. number of message snippets to retrieve

        Example:
            Fetch the latest message in the thread that matches the query.

            >>> (message,) = thread.search_messages("abc", limit=1)
            >>> message.text
            "Some text and abc"
        """
        offset = 0
        # The max limit is measured empirically to 420, safe default chosen below
        for limit in _util.get_limits(limit, max_limit=50):
            _, snippets = await self._search_messages(query, offset, limit)
            for snippet in snippets:
                yield snippet
            if len(snippets) < limit:
                return  # No more data to fetch
            offset += limit

    async def _fetch_messages(self, limit, before):
        params = {
            "id": self.id,
            "message_limit": limit,
            "load_messages": True,
            "load_read_receipts": True,
            # "load_delivery_receipts": False,
            # "is_work_teamwork_not_putting_muted_in_unreads": False,
            "before": _util.datetime_to_millis(before) if before else None,
        }
        (j,) = await self.session._graphql_requests(
            _graphql.from_doc_id("1860982147341344", params)  # 2696825200377124
        )

        if j.get("message_thread") is None:
            raise _exception.ParseError("Could not fetch messages", data=j)

        # TODO: Should we parse the returned thread data, too?

        read_receipts = j["message_thread"]["read_receipts"]["nodes"]

        thread = self._copy()
        return [
            _models.MessageData._from_graphql(thread, message, read_receipts)
            for message in j["message_thread"]["messages"]["nodes"]
        ]

    async def fetch_messages(self, limit: Optional[int]
                             ) -> AsyncGenerator["_models.MessageData", None]:
        """Fetch messages in a thread.

        The returned messages are ordered by last sent first.

        Args:
            limit: Max. number of threads to retrieve. If ``None``, all threads will be
                retrieved.

        Example:
            >>> for message in thread.fetch_messages(limit=5)
            ...     print(message.text)
            ...
            A message
            Another message
            None
            A fourth message
        """
        # This is measured empirically as 210 in extreme cases, fairly safe default
        # chosen below
        MAX_BATCH_LIMIT = 100

        before = None
        for limit in _util.get_limits(limit, MAX_BATCH_LIMIT):
            messages = await self._fetch_messages(limit, before)
            messages.reverse()

            if before:
                # Strip the first messages
                for message in messages[1:]:
                    yield message
            else:
                for message in messages:
                    yield message

            if len(messages) < MAX_BATCH_LIMIT:
                return  # No more data to fetch

            before = messages[-1].created_at

    async def _fetch_images(self, limit, after):
        data = {"id": self.id, "first": limit, "after": after}
        (j,) = await self.session._graphql_requests(
            _graphql.from_query_id("515216185516880", data)
        )

        if not j[self.id]:
            raise _exception.ParseError("Could not find images", data=j)

        result = j[self.id]["message_shared_media"]

        rtn = []
        for edge in result["edges"]:
            node = edge["node"]
            type_ = node["__typename"]
            if type_ == "MessageImage":
                rtn.append(_models.ImageAttachment._from_list(node))
            elif type_ == "MessageVideo":
                rtn.append(_models.VideoAttachment._from_list(node))
            else:
                log.warning("Unknown image type %s, data: %s", type_, edge)
                rtn.append(None)

        # result["page_info"]["has_next_page"] is not correct when limit > 12
        return (result["page_info"]["end_cursor"], rtn)

    async def fetch_images(self, limit: Optional[int]) -> Iterable["_models.Attachment"]:
        """Fetch images/videos posted in the thread.

        The returned images are ordered by last sent first.

        Args:
            limit: Max. number of images to retrieve. If ``None``, all images will be
                retrieved.

        Example:
            >>> for image in thread.fetch_messages(limit=3)
            ...     print(image.id)
            ...
            1234
            2345
        """
        cursor = None
        # The max limit on this request is unknown, so we set it reasonably high
        # This way `limit=None` also still works
        for limit in _util.get_limits(limit, max_limit=1000):
            cursor, images = await self._fetch_images(limit, cursor)
            if not images:
                return  # No more data to fetch
            for image in images:
                if image:
                    yield image

    async def set_nickname(self, user_id: str, nickname: str):
        """Change the nickname of a user in the thread.

        Args:
            user_id: User that will have their nickname changed
            nickname: New nickname

        Example:
            >>> thread.set_nickname("1234", "A nickname")
        """
        data = {
            "nickname": nickname,
            "participant_id": user_id,
            "thread_or_other_fbid": self.id,
        }
        j = await self.session._payload_post(
            "/messaging/save_thread_nickname/?source=thread_settings&dpr=1", data
        )

    async def set_color(self, color: str):
        """Change thread color.

        The new color must be one of the following::

            "#0084ff", "#44bec7", "#ffc300", "#fa3c4c", "#d696bb", "#6699cc",
            "#13cf13", "#ff7e29", "#e68585", "#7646ff", "#20cef5", "#67b868",
            "#d4a88c", "#ff5ca1", "#a695c7", "#ff7ca8", "#1adb5b", "#f01d6a",
            "#ff9c19" or "#0edcde".

        This list is subject to change in the future!

        The default when creating a new thread is ``"#0084ff"``.

        Args:
            color: New thread color

        Example:
            Set the thread color to "Coral Pink".

            >>> thread.set_color("#e68585")
        """
        if color not in SETABLE_COLORS:
            raise ValueError(
                "Invalid color! Please use one of: {}".format(SETABLE_COLORS)
            )

        # Set color to "" if DEFAULT_COLOR. Just how the endpoint works...
        if color == DEFAULT_COLOR:
            color = ""

        data = {"color_choice": color, "thread_or_other_fbid": self.id}
        j = await self.session._payload_post(
            "/messaging/save_thread_color/?source=thread_settings&dpr=1", data
        )

    # async def set_theme(self, theme_id: str):
    #     data = {
    #         "client_mutation_id": "0",
    #         "actor_id": self.session.user.id,
    #         "thread_id": self.id,
    #         "theme_id": theme_id,
    #         "source": "SETTINGS",
    #     }
    #     j = await self.session._graphql_requests(
    #         _graphql.from_doc_id("1768656253222505", {"data": data})
    #     )

    async def set_emoji(self, emoji: Optional[str]):
        """Change thread emoji.

        Args:
            emoji: New thread emoji. If ``None``, will be set to the default "LIKE" icon

        Example:
            Set the thread emoji to "😊".

            >>> thread.set_emoji("😊")
        """
        data = {"emoji_choice": emoji, "thread_or_other_fbid": self.id}
        # While changing the emoji, the Facebook web client actually sends multiple
        # different requests, though only this one is required to make the change.
        j = await self.session._payload_post(
            "/messaging/save_thread_emoji/?source=thread_settings&dpr=1", data
        )

    async def forward_attachment(self, attachment_id: str):
        """Forward an attachment.

        Args:
            attachment_id: Attachment ID to forward

        Example:
            >>> thread.forward_attachment("1234")
        """
        data = {
            "attachment_id": attachment_id,
            "recipient_map[{}]".format(_util.generate_offline_threading_id()): self.id,
        }
        j = await self.session._payload_post("/mercury/attachments/forward/", data)
        if not j.get("success"):
            raise _exception.ExternalError("Failed forwarding attachment", j["error"])

    async def _set_typing(self, typing):
        data = {
            "typ": "1" if typing else "0",
            "thread": self.id,
            # TODO: This
            # "to": self.id if isinstance(self, _user.User) else "",
            "source": "mercury-chat",
        }
        j = await self.session._payload_post("/ajax/messaging/typ.php", data)

    async def start_typing(self):
        """Set the current user to start typing in the thread.

        Example:
            >>> thread.start_typing()
        """
        await self._set_typing(True)

    async def stop_typing(self):
        """Set the current user to stop typing in the thread.

        Example:
            >>> thread.stop_typing()
        """
        await self._set_typing(False)

    async def create_plan(
        self,
        name: str,
        at: datetime.datetime,
        location_name: str = None,
        location_id: str = None,
    ):
        """Create a new plan.

        # TODO: Arguments

        Args:
            name: Name of the new plan
            at: When the plan is for

        Example:
            >>> thread.create_plan(...)
        """
        return await _models.Plan._create(self, name, at, location_name, location_id)

    async def create_poll(self, question: str, options: Mapping[str, bool]):
        """Create poll in a thread.

        Args:
            question: The question
            options: Options and whether you want to select the option

        Example:
            >>> thread.create_poll("Test poll", {"Option 1": True, "Option 2": False})
        """
        # We're using ordered dictionaries, because the Facebook endpoint that parses
        # the POST parameters is badly implemented, and deals with ordering the options
        # wrongly. If you can find a way to fix this for the endpoint, or if you find
        # another endpoint, please do suggest it ;)
        data = collections.OrderedDict(
            [("question_text", question), ("target_id", self.id)]
        )

        for i, (text, vote) in enumerate(options.items()):
            data["option_text_array[{}]".format(i)] = text
            data["option_is_selected_array[{}]".format(i)] = "1" if vote else "0"

        j = await self.session._payload_post(
            "/messaging/group_polling/create_poll/?dpr=1", data
        )
        if j.get("status") != "success":
            raise _exception.ExternalError(
                "Failed creating poll: {}".format(j.get("errorTitle")),
                j.get("errorMessage"),
            )

    async def mute(self, duration: datetime.timedelta = None):
        """Mute the thread.

        Args:
            duration: Time to mute, use ``None`` to mute forever

        Example:
            >>> import datetime
            >>> thread.mute(datetime.timedelta(days=2))
        """
        if duration is None:
            setting = "-1"
        else:
            setting = str(_util.timedelta_to_seconds(duration))
        data = {"mute_settings": setting, "thread_fbid": self.id}
        j = await self.session._payload_post(
            "/ajax/mercury/change_mute_thread.php?dpr=1", data
        )

    async def unmute(self):
        """Unmute the thread.

        Example:
            >>> thread.unmute()
        """
        return await self.mute(datetime.timedelta(0))

    async def _mute_reactions(self, mode: bool):
        data = {"reactions_mute_mode": "1" if mode else "0", "thread_fbid": self.id}
        j = await self.session._payload_post(
            "/ajax/mercury/change_reactions_mute_thread/?dpr=1", data
        )

    async def mute_reactions(self):
        """Mute thread reactions."""
        await self._mute_reactions(True)

    async def unmute_reactions(self):
        """Unmute thread reactions."""
        await self._mute_reactions(False)

    async def _mute_mentions(self, mode: bool):
        data = {"mentions_mute_mode": "1" if mode else "0", "thread_fbid": self.id}
        j = await self.session._payload_post(
            "/ajax/mercury/change_mentions_mute_thread/?dpr=1", data
        )

    async def mute_mentions(self):
        """Mute thread mentions."""
        await self._mute_mentions(True)

    async def unmute_mentions(self):
        """Unmute thread mentions."""
        await self._mute_mentions(False)

    async def mark_as_spam(self):
        """Mark the thread as spam, and delete it."""
        data = {"id": self.id}
        j = await self.session._payload_post("/ajax/mercury/mark_spam.php?dpr=1", data)

    @staticmethod
    async def _delete_many(session, thread_ids):
        data = {}
        for i, id_ in enumerate(thread_ids):
            data["ids[{}]".format(i)] = id_
        # Not needed any more
        # j = session._payload_post("/ajax/mercury/change_pinned_status.php?dpr=1", ...)
        # Both /ajax/mercury/delete_threads.php (with an s) doesn't work
        j = await session._payload_post("/ajax/mercury/delete_thread.php", data)

    async def delete(self):
        """Delete the thread.

        If you want to delete multiple threads, please use `Client.delete_threads`.

        Example:
            >>> message.delete()
        """
        await self._delete_many(self.session, [self.id])

    async def _forced_fetch(self, message_id: str) -> dict:
        params = {
            "thread_and_message_id": {"thread_id": self.id, "message_id": message_id}
        }
        (j,) = await self.session._graphql_requests(
            _graphql.from_doc_id("1768656253222505", params)
        )
        return j

    @staticmethod
    def _parse_color(inp: Optional[str]) -> str:
        if not inp:
            return DEFAULT_COLOR
        # Strip the alpha value, and lower the string
        return "#{}".format(inp[2:].lower())

    @staticmethod
    def _parse_customization_info(data: Any) -> MutableMapping[str, Any]:
        if not data or not data.get("customization_info"):
            return {"emoji": None, "color": DEFAULT_COLOR}
        info = data["customization_info"]

        rtn = {
            "emoji": info.get("emoji"),
            "color": ThreadABC._parse_color(info.get("outgoing_bubble_color")),
        }
        if (
            data.get("thread_type") == "GROUP"
            or data.get("is_group_thread")
            or data.get("thread_key", {}).get("thread_fbid")
        ):
            rtn["nicknames"] = {}
            for k in info.get("participant_customizations", []):
                rtn["nicknames"][k["participant_id"]] = k.get("nickname")
        elif info.get("participant_customizations"):
            user_id = data.get("thread_key", {}).get("other_user_id") or data.get("id")
            pc = info["participant_customizations"]
            if len(pc) > 0:
                if pc[0].get("participant_id") == user_id:
                    rtn["nickname"] = pc[0].get("nickname")
                else:
                    rtn["own_nickname"] = pc[0].get("nickname")
            if len(pc) > 1:
                if pc[1].get("participant_id") == user_id:
                    rtn["nickname"] = pc[1].get("nickname")
                else:
                    rtn["own_nickname"] = pc[1].get("nickname")
        return rtn

    @staticmethod
    def _parse_participants(session, data) -> Iterable["ThreadABC"]:
        from . import _user, _group, _page

        for node in data["nodes"]:
            actor = node["messaging_actor"]
            typename = actor["__typename"]
            thread_id = actor["id"]
            if typename == "User":
                yield _user.User(session=session, id=thread_id)
            elif typename == "MessageThread":
                # MessageThread => Group thread
                yield _group.Group(session=session, id=thread_id)
            elif typename == "Page":
                yield _page.Page(session=session, id=thread_id)
            elif typename == "Group":
                # We don't handle Facebook "Groups"
                pass
            else:
                log.warning("Unknown type %r in %s", typename, data)


@attrs_default
class Thread(ThreadABC):
    """Represents a Facebook thread, where the actual type is unknown.

    Implements parts of `ThreadABC`, call the method to figure out if your use case is
    supported. Otherwise, you'll have to use an `User`/`Group`/`Page` object.

    Note: This list may change in minor versions!
    """

    #: The session to use when making requests.
    session: _session.Session
    #: The unique identifier of the thread.
    id: str = attr.ib(converter=str)

    def _to_send_data(self):
        raise NotImplementedError(
            "The method you called is not supported on raw Thread objects."
            " Please use an appropriate User/Group/Page object instead!"
        )

    def _copy(self) -> "Thread":
        return Thread(session=self.session, id=self.id)



import attr
from . import Image
from .._common import attrs_default
from .. import _util

from typing import Optional, Sequence


@attrs_default
class Attachment:
    """Represents a Facebook attachment."""

    #: The attachment ID
    id: Optional[str] = None


@attrs_default
class UnsentMessage(Attachment):
    """Represents an unsent message attachment."""


@attrs_default
class ShareAttachment(Attachment):
    """Represents a shared item (e.g. URL) attachment."""

    #: ID of the author of the shared post
    author: Optional[str] = None
    #: Target URL
    url: Optional[str] = None
    #: Original URL if Facebook redirects the URL
    original_url: Optional[str] = None
    #: Title of the attachment
    title: Optional[str] = None
    #: Description of the attachment
    description: Optional[str] = None
    #: Name of the source
    source: Optional[str] = None
    #: The attached image
    image: Optional[Image] = None
    #: URL of the original image if Facebook uses ``safe_image``
    original_image_url: Optional[str] = None
    #: List of additional attachments
    attachments: Sequence[Attachment] = attr.ib(factory=list)

    @classmethod
    def _from_graphql(cls, data):
        from . import _file

        image = None
        original_image_url = None
        media = data.get("media")
        if media and media.get("image"):
            image = Image._from_uri(media["image"])
            original_image_url = (
                _util.get_url_parameter(image.url, "url")
                if "/safe_image.php" in image.url
                else image.url
            )

        url = data.get("url")
        return cls(
            id=data.get("deduplication_key"),
            author=data["target"]["actors"][0]["id"]
            if data["target"].get("actors")
            else None,
            url=url,
            original_url=_util.get_url_parameter(url, "u")
            if "/l.php?u=" in url
            else url,
            title=data["title_with_entities"].get("text"),
            description=data["description"].get("text")
            if data.get("description")
            else None,
            source=data["source"].get("text") if data.get("source") else None,
            image=image,
            original_image_url=original_image_url,
            attachments=[
                _file.graphql_to_subattachment(attachment)
                for attachment in data.get("subattachments")
            ],
        )



"""Everything in this module is taken from the excellent trio project.

Having the public path in .__module__ attributes is important for:
- exception names in printed tracebacks
- ~sphinx :show-inheritance:~
- deprecation warnings
- pickle
- probably other stuff
"""

import os


def fixup_module_metadata(namespace):
    def fix_one(qualname, name, obj):
        # Custom extension, to handle classmethods, staticmethods and properties
        if isinstance(obj, (classmethod, staticmethod)):
            obj = obj.__func__
        if isinstance(obj, property):
            obj = obj.fget

        mod = getattr(obj, "__module__", None)
        if mod is not None and mod.startswith("fbchat."):
            obj.__module__ = "fbchat"
            # Modules, unlike everything else in Python, put fully-qualitied
            # names into their __name__ attribute. We check for "." to avoid
            # rewriting these.
            if hasattr(obj, "__name__") and "." not in obj.__name__:
                obj.__name__ = name
                obj.__qualname__ = qualname
            if isinstance(obj, type):
                # Fix methods
                for attr_name, attr_value in obj.__dict__.items():
                    fix_one(objname + "." + attr_name, attr_name, attr_value)

    for objname, obj in namespace.items():
        if not objname.startswith("_"):  # ignore private attributes
            fix_one(objname, objname, obj)


# Allow disabling this when running Sphinx
# This is done so that Sphinx autodoc can detect the file's source
# TODO: Find a better way to detect when we're running Sphinx!
if os.environ.get("_FBCHAT_DISABLE_FIX_MODULE_METADATA") == "1":
    fixup_module_metadata = lambda namespace: None



import attr
import datetime
from ._common import attrs_event, Event, UnknownEvent, ThreadEvent
from . import _delta_type
from .. import _util, _threads, _models

from typing import Sequence, Optional


@attrs_event
class PeopleAdded(ThreadEvent):
    """somebody added people to a group thread."""

    # TODO: Add message id

    thread: "_threads.Group"  # Set the correct type
    #: The people who got added
    added: Sequence["_threads.User"]
    #: When the people were added
    at: datetime.datetime

    @classmethod
    def _parse(cls, session, data):
        author, thread, at = cls._parse_metadata(session, data)
        added = [
            # TODO: Parse user name
            _threads.User(session=session, id=x["userFbId"])
            for x in data["addedParticipants"]
        ]
        return cls(author=author, thread=thread, added=added, at=at)


@attrs_event
class PersonRemoved(ThreadEvent):
    """Somebody removed a person from a group thread."""

    # TODO: Add message id

    thread: "_threads.Group"  # Set the correct type
    #: Person who got removed
    removed: "_threads.User"
    #: When the person were removed
    at: datetime.datetime

    @classmethod
    def _parse(cls, session, data):
        author, thread, at = cls._parse_metadata(session, data)
        removed = _threads.User(session=session, id=data["leftParticipantFbId"])
        return cls(author=author, thread=thread, removed=removed, at=at)


@attrs_event
class TitleSet(ThreadEvent):
    """Somebody changed a group's title."""

    thread: "_threads.Group"  # Set the correct type
    #: The new title. If ``None``, the title was removed
    title: Optional[str]
    #: When the title was set
    at: datetime.datetime

    @classmethod
    def _parse(cls, session, data):
        author, thread, at = cls._parse_metadata(session, data)
        return cls(author=author, thread=thread, title=data["name"] or None, at=at)


@attrs_event
class UnfetchedThreadEvent(Event):
    """A message was received, but the data must be fetched manually.

    Use `Message.fetch` to retrieve the message data.

    This is usually used when somebody changes the group's photo, or when a new pending
    group is created.
    """

    # TODO: Present this in a way that users can fetch the changed group photo easily

    #: The thread the message was sent to
    thread: "_threads.ThreadABC"
    #: The message
    message: Optional["_models.Message"]

    @classmethod
    def _parse(cls, session, data):
        thread = cls._get_thread(session, data)
        message = None
        if "messageId" in data:
            message = _models.Message(thread=thread, id=data["messageId"])
        return cls(thread=thread, message=message)


@attrs_event
class MessagesDelivered(ThreadEvent):
    """Somebody marked messages as delivered in a thread."""

    #: The messages that were marked as delivered
    messages: Sequence["_models.Message"]
    #: When the messages were delivered
    at: datetime.datetime

    @classmethod
    def _parse(cls, session, data):
        thread = cls._get_thread(session, data)
        if "actorFbId" in data:
            author = _threads.User(session=session, id=data["actorFbId"])
        else:
            author = thread
        messages = [_models.Message(thread=thread, id=x) for x in data["messageIds"]]
        at = _util.millis_to_datetime(int(data["deliveredWatermarkTimestampMs"]))
        return cls(author=author, thread=thread, messages=messages, at=at)


@attrs_event
class ThreadsRead(Event):
    """Somebody marked threads as read/seen."""

    #: The person who marked the threads as read
    author: "_threads.ThreadABC"
    #: The threads that were marked as read
    threads: Sequence["_threads.ThreadABC"]
    #: When the threads were read
    at: datetime.datetime

    @classmethod
    def _parse_read_receipt(cls, session, data):
        author = _threads.User(session=session, id=data["actorFbId"])
        thread = cls._get_thread(session, data)
        at = _util.millis_to_datetime(int(data["actionTimestampMs"]))
        return cls(author=author, threads=[thread], at=at)

    @classmethod
    def _parse(cls, session, data):
        threads = [
            cls._get_thread(session, {"threadKey": x}) for x in data["threadKeys"]
        ]
        at = _util.millis_to_datetime(int(data["actionTimestamp"]))
        return cls(author=session.user, threads=threads, at=at)


@attrs_event
class MessageEvent(ThreadEvent):
    """Somebody sent a message to a thread."""

    #: The sent message
    message: "_models.MessageData"
    #: When the threads were read
    at: datetime.datetime

    @classmethod
    def _parse(cls, session, data):
        author, thread, at = cls._parse_metadata(session, data)
        message = _models.MessageData._from_pull(
            thread, data, author=author.id, created_at=at,
        )
        return cls(author=author, thread=thread, message=message, at=at)


@attrs_event
class ThreadFolder(Event):
    """A thread was created in a folder.

    Somebody that isn't connected with you on either Facebook or Messenger sends a
    message. After that, you need to use `ThreadABC.fetch_messages` to actually read it.
    """

    # TODO: Finish this

    #: The created thread
    thread: "_threads.ThreadABC"
    #: The folder/location
    folder: "_models.ThreadLocation"

    @classmethod
    def _parse(cls, session, data):
        thread = cls._get_thread(session, data)
        folder = _models.ThreadLocation._parse(data["folder"])
        return cls(thread=thread, folder=folder)


def parse_delta(session, data):
    class_ = data["class"]
    if class_ == "AdminTextMessage":
        return _delta_type.parse_admin_message(session, data)
    elif class_ == "ParticipantsAddedToGroupThread":
        return PeopleAdded._parse(session, data)
    elif class_ == "ParticipantLeftGroupThread":
        return PersonRemoved._parse(session, data)
    elif class_ == "MarkFolderSeen":
        # TODO: Finish this
        folders = [_models.ThreadLocation._parse(folder) for folder in data["folders"]]
        at = _util.millis_to_datetime(int(data["timestamp"]))
        return None
    elif class_ == "ThreadName":
        return TitleSet._parse(session, data)
    elif class_ == "ForcedFetch":
        return UnfetchedThreadEvent._parse(session, data)
    elif class_ == "DeliveryReceipt":
        return MessagesDelivered._parse(session, data)
    elif class_ == "ReadReceipt":
        return ThreadsRead._parse_read_receipt(session, data)
    elif class_ == "MarkRead":
        return ThreadsRead._parse(session, data)
    elif class_ == "NoOp":
        # Skip "no operation" events
        return None
    elif class_ == "NewMessage":
        return MessageEvent._parse(session, data)
    elif class_ == "ThreadFolder":
        return ThreadFolder._parse(session, data)
    elif class_ == "ClientPayload":
        raise ValueError("This is implemented in `parse_events`")
    return UnknownEvent(source="Delta class", data=data)



import setuptools

try:
    long_desc = open("README.rst").read()
except IOError:
    long_desc = "Failed to read README.rst"

setuptools.setup(
    name="fbchat-asyncio",
    version="0.6.0b1",
    url="https://github.com/tulir/fbchat-asyncio",

    author="Tulir Asokan",
    author_email="tulir@maunium.net",

    description="Facebook Messenger library for Python/Asyncio.",
    long_description=long_desc,

    packages=setuptools.find_packages(),

    install_requires=[
        "attrs>=19.1",
        "beautifulsoup4~=4.0",
        "aiohttp",
        "yarl",
        "paho-mqtt~=1.5",
    ],

    python_requires="~=3.6",

    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Intended Audience :: Information Technology",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Natural Language :: English",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Framework :: AsyncIO",
        "Topic :: Communications :: Chat",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)



import attr
import datetime
import enum
from .._common import attrs_default
from .. import _exception, _util, _session

from typing import Mapping, Sequence, Optional


class GuestStatus(enum.Enum):
    INVITED = 1
    GOING = 2
    DECLINED = 3


ACONTEXT = {
    "action_history": [
        {"surface": "messenger_chat_tab", "mechanism": "messenger_composer"}
    ]
}


@attrs_default
class Plan:
    """Base model for plans.

    Example:
        >>> plan = fbchat.Plan(session=session, id="1234")
    """

    #: The session to use when making requests.
    session: _session.Session
    #: The plan's unique identifier.
    id: str = attr.ib(converter=str)

    async def fetch(self) -> "PlanData":
        """Fetch fresh `PlanData` object.

        Example:
            >>> plan = plan.fetch()
            >>> plan.title
            "A plan"
        """
        data = {"event_reminder_id": self.id}
        j = await self.session._payload_post("/ajax/eventreminder", data)
        return PlanData._from_fetch(self.session, j)

    @classmethod
    async def _create(
        cls,
        thread,
        name: str,
        at: datetime.datetime,
        location_name: str = None,
        location_id: str = None,
    ):
        data = {
            "event_type": "EVENT",
            "event_time": _util.datetime_to_seconds(at),
            "title": name,
            "thread_id": thread.id,
            "location_id": location_id or "",
            "location_name": location_name or "",
            "acontext": ACONTEXT,
        }
        j = await thread.session._payload_post("/ajax/eventreminder/create", data)
        if "error" in j:
            raise _exception.ExternalError("Failed creating plan", j["error"])

    async def edit(
        self,
        name: str,
        at: datetime.datetime,
        location_name: str = None,
        location_id: str = None,
    ):
        """Edit the plan.

        # TODO: Arguments
        """
        data = {
            "event_reminder_id": self.id,
            "delete": "false",
            "date": _util.datetime_to_seconds(at),
            "location_name": location_name or "",
            "location_id": location_id or "",
            "title": name,
            "acontext": ACONTEXT,
        }
        j = await self.session._payload_post("/ajax/eventreminder/submit", data)

    async def delete(self):
        """Delete the plan.

        Example:
            >>> plan.delete()
        """
        data = {"event_reminder_id": self.id, "delete": "true", "acontext": ACONTEXT}
        j = await self.session._payload_post("/ajax/eventreminder/submit", data)

    async def _change_participation(self, take_part):
        data = {
            "event_reminder_id": self.id,
            "guest_state": "GOING" if take_part else "DECLINED",
            "acontext": ACONTEXT,
        }
        j = await self.session._payload_post("/ajax/eventreminder/rsvp", data)

    async def participate(self):
        """Set yourself as GOING/participating to the plan.

        Example:
            >>> plan.participate()
        """
        return await self._change_participation(True)

    async def decline(self):
        """Set yourself as having DECLINED the plan.

        Example:
            >>> plan.decline()
        """
        return await self._change_participation(False)


@attrs_default
class PlanData(Plan):
    """Represents data about a plan."""

    #: Plan time, only precise down to the minute
    time: datetime.datetime
    #: Plan title
    title: str
    #: Plan location name
    location: Optional[str] = attr.ib(default=None, converter=lambda x: x or "")
    #: Plan location ID
    location_id: Optional[str] = attr.ib(default=None, converter=lambda x: x or "")
    #: ID of the plan creator
    author_id: Optional[str] = None
    #: `User` ids mapped to their `GuestStatus`
    guests: Optional[Mapping[str, GuestStatus]] = None

    @property
    def going(self) -> Sequence[str]:
        """List of the `User` IDs who will take part in the plan."""
        return [
            id_
            for id_, status in (self.guests or {}).items()
            if status is GuestStatus.GOING
        ]

    @property
    def declined(self) -> Sequence[str]:
        """List of the `User` IDs who won't take part in the plan."""
        return [
            id_
            for id_, status in (self.guests or {}).items()
            if status is GuestStatus.DECLINED
        ]

    @property
    def invited(self) -> Sequence[str]:
        """List of the `User` IDs who are invited to the plan."""
        return [
            id_
            for id_, status in (self.guests or {}).items()
            if status is GuestStatus.INVITED
        ]

    @classmethod
    def _from_pull(cls, session, data):
        return cls(
            session=session,
            id=data.get("event_id"),
            time=_util.seconds_to_datetime(int(data.get("event_time"))),
            title=data.get("event_title"),
            location=data.get("event_location_name"),
            location_id=data.get("event_location_id"),
            author_id=data.get("event_creator_id"),
            guests={
                x["node"]["id"]: GuestStatus[x["guest_list_state"]]
                for x in _util.parse_json(data["guest_state_list"])
            },
        )

    @classmethod
    def _from_fetch(cls, session, data):
        return cls(
            session=session,
            id=data.get("oid"),
            time=_util.seconds_to_datetime(data.get("event_time")),
            title=data.get("title"),
            location=data.get("location_name"),
            location_id=str(data["location_id"]) if data.get("location_id") else None,
            author_id=data.get("creator_id"),
            guests={id_: GuestStatus[s] for id_, s in data["event_members"].items()},
        )

    @classmethod
    def _from_graphql(cls, session, data):
        return cls(
            session=session,
            id=data.get("id"),
            time=_util.seconds_to_datetime(data.get("time")),
            title=data.get("event_title"),
            location=data.get("location_name"),
            author_id=data["lightweight_event_creator"].get("id"),
            guests={
                x["node"]["id"]: GuestStatus[x["guest_list_state"]]
                for x in data["event_reminder_members"]["edges"]
            },
        )
